"""In this test module, we test the torch spherical harmonics module."""

import pytest
import torch
import torch_harmonics as harmonics

from fme.core.device import get_device


@pytest.mark.parametrize("nlat, nlon", [(6, 12)])
@pytest.mark.parametrize("grid", ["equiangular", "legendre-gauss"])
@pytest.mark.parametrize("constant", [1.0, 0.42])
def test_constant_field(nlat, nlon, grid, constant):
    """Tests that the SHT of a constant field has a single non-zero wavenumber,
    the first one.
    """
    constant_field = torch.tensor(constant, device=get_device()).repeat(nlat, nlon)

    sht = harmonics.RealSHT(nlat, nlon, grid=grid).to(get_device())

    coeffs = sht(constant_field).ravel()
    zero = torch.zeros(1, dtype=torch.complex64, device=get_device())

    assert not torch.isclose(coeffs[0], zero)
    assert torch.all(torch.isclose(zero, coeffs[1:], atol=1e-6))


def _roundtrip(field, grid):
    nlat, nlon = field.shape
    sht = harmonics.RealSHT(nlat, nlon, grid=grid).to(get_device())
    isht = harmonics.InverseRealSHT(nlat, nlon, grid=grid).to(get_device())
    return isht(sht(field))


@pytest.mark.parametrize("seed", [0, 1])
@pytest.mark.parametrize("nlat, nlon", [(6, 12)])
def test_roundtrip(nlat, nlon, seed, grid="legendre-gauss"):
    """Tests that the SHT and ISHT are inverses of each other."""
    torch.manual_seed(seed)
    random_field = torch.randn(nlat, nlon, device=get_device())
    proj = _roundtrip(random_field, grid)
    assert torch.all(torch.isclose(proj, _roundtrip(proj, grid), atol=1e-6))



# import modules so they are registered
from . import m2lines as _m2lines
from . import prebuilt as _prebuilt
from . import sfno as _sfno
from . import stochastic_sfno as _sfno_crps
from .registry import ModuleSelector

del _prebuilt, _sfno, _m2lines



import copy
from collections.abc import Mapping
from pathlib import Path

import numpy as np
import torch
import xarray as xr

from fme.ace.aggregator.inference.video import VideoAggregator
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.core.dataset.data_typing import VariableMetadata


class PairedVideoDataWriter:
    """
    Write [time, lat, lon] metric data to a netCDF file.
    """

    def __init__(
        self,
        path: str,
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
        label: str = "reduced_autoregressive_predictions",
    ):
        """
        Args:
            path: Directory within which to write the file.
            n_samples: Number of samples to write to the file.
            n_timesteps: Number of timesteps to write to the file.
            variable_metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
            dataset_metadata: Metadata for the dataset.
            label: Label for the filename.
        """
        self.path = path
        self._metrics_filename = str(Path(path) / f"{label}.nc")
        self.variable_metadata = variable_metadata
        self.coords = coords
        self._video = VideoAggregator(
            n_timesteps=n_timesteps,
            enable_extended_videos=True,
            variable_metadata=variable_metadata,
        )
        self._dataset_metadata = copy.copy(dataset_metadata)
        self._dataset_metadata.title = f"ACE {label.replace('_', ' ')} data file"

    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            target: Target data.
            prediction: Prediction data.
            start_timestep: Timestep at which to start writing.
            batch_time: Time coordinate for each sample in the batch. Unused.
        """
        self._video.record_batch(
            target_data=target,
            gen_data=prediction,
            i_time_start=start_timestep,
        )

    def flush(self):
        """
        Flush the data to disk.
        """
        metric_dataset = self._video.get_dataset()
        coords = {}
        if "lat" in self.coords:
            coords["lat"] = self.coords["lat"]
        if "lon" in self.coords:
            coords["lon"] = self.coords["lon"]
        metric_dataset = metric_dataset.assign_coords(coords)
        metric_dataset.attrs.update(self._dataset_metadata.as_flat_str_dict())
        metric_dataset.to_netcdf(self._metrics_filename)



import torch.optim.lr_scheduler

from fme.core.scheduler import SchedulerConfig


def test_default_gives_none():
    optimizer = torch.optim.Adam(params=[torch.nn.Parameter()])
    max_epochs = 42
    assert SchedulerConfig().build(optimizer, max_epochs) is None


def test_build():
    config = SchedulerConfig(type="StepLR", kwargs={"step_size": 1})
    # define dummy parameters for optimizer
    optimizer = torch.optim.Adam(params=[torch.nn.Parameter()])
    max_epochs = 42
    scheduler = config.build(optimizer, max_epochs)
    assert isinstance(scheduler, torch.optim.lr_scheduler.StepLR)
    assert scheduler.step_size == 1
    assert scheduler.optimizer is optimizer



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import List, Optional, Sequence

import torch.nn as nn

from fme.ace.models.healpix.healpix_activations import DownsamplingBlockConfig

from .healpix_blocks import ConvBlockConfig


@dataclasses.dataclass
class UNetEncoderConfig:
    """
    Configuration for the UNet Encoder.

    Parameters:
        conv_block: Configuration for the convolutional block.
        down_sampling_block: Configuration for the down-sampling block.
        input_channels: Number of input channels, by default 3.
        n_channels: Number of channels for each layer, by default (136, 68, 34).
        n_layers: Number of layers in each block, by default (2, 2, 1).
        dilations: List of dilation rates for the layers, by default None.
        enable_nhwc: Flag to enable NHWC data format, by default False.
        enable_healpixpad: Flag to enable HEALPix padding, by default False.
    """

    conv_block: ConvBlockConfig
    down_sampling_block: DownsamplingBlockConfig
    input_channels: int = 3
    n_channels: List[int] = dataclasses.field(default_factory=lambda: [136, 68, 34])
    n_layers: List[int] = dataclasses.field(default_factory=lambda: [2, 2, 1])
    dilations: Optional[list] = None
    enable_nhwc: bool = False
    enable_healpixpad: bool = False

    def build(self) -> nn.Module:
        """
        Builds the UNet Encoder model.

        Returns:
            UNet Encoder model.
        """
        return UNetEncoder(
            conv_block=self.conv_block,
            down_sampling_block=self.down_sampling_block,
            input_channels=self.input_channels,
            n_channels=self.n_channels,
            n_layers=self.n_layers,
            dilations=self.dilations,
            enable_nhwc=self.enable_nhwc,
            enable_healpixpad=self.enable_healpixpad,
        )


class UNetEncoder(nn.Module):
    """Generic UNetEncoder that can be applied to arbitrary meshes."""

    def __init__(
        self,
        conv_block: ConvBlockConfig,
        down_sampling_block: DownsamplingBlockConfig,
        input_channels: int = 3,
        n_channels: Sequence = (16, 32, 64),
        n_layers: Sequence = (2, 2, 1),
        dilations: Optional[list] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Args:
            conv_block: config for the convolutional block
            down_sampling_block: DownsamplingBlockConfig for the downsample block
            input_channels: # of input channels
            n_channels: # of channels in each encoder layer
            n_layers:, # of layers to use for the convolutional blocks
            dilations: list of dilations to use for the the convolutional blocks
            enable_nhwc: if channel last format should be used
            enable_healpixpad: if healpixpad library should be used (true if installed)
        """
        super().__init__()
        self.n_channels = n_channels

        if dilations is None:
            # Defaults to [1, 1, 1...] in accordance with the number of unet levels
            dilations = [1 for _ in range(len(n_channels))]

        # Build encoder
        old_channels = input_channels
        self.encoder = []
        for n, curr_channel in enumerate(n_channels):
            modules = list()
            if n > 0:
                down_sampling_block.enable_nhwc = enable_nhwc
                down_sampling_block.enable_healpixpad = enable_healpixpad
                modules.append(
                    down_sampling_block.build()  # Shapes are not used in these calls.
                )

            # Set up conv block
            conv_block.in_channels = old_channels
            conv_block.latent_channels = curr_channel
            conv_block.out_channels = curr_channel
            conv_block.dilation = dilations[n]
            conv_block.n_layers = n_layers[n]
            conv_block.enable_nhwc = enable_nhwc
            conv_block.enable_healpixpad = enable_healpixpad
            modules.append(conv_block.build())  # Shapes are not used in these calls.
            old_channels = curr_channel

            self.encoder.append(nn.Sequential(*modules))

        self.encoder = nn.ModuleList(self.encoder)

    def forward(self, inputs: Sequence) -> Sequence:
        """
        Forward pass of the HEALPix Unet encoder

        Args:
            inputs: The inputs to enccode

        Returns:
            The encoded values
        """
        outputs = []
        for layer in self.encoder:
            outputs.append(layer(inputs))
            inputs = outputs[-1]
        return outputs

    def reset(self):
        """Resets the state of the decoder layers"""
        pass



import copy
from typing import Literal

import pytest
import torch
import torch.nn as nn
import torch.random
import yaml

import fme
from fme.core.optimization import (
    Checkpoint,
    CheckpointConfig,
    NoCheckpoint,
    NullOptimization,
    Optimization,
)
from fme.core.scheduler import SchedulerConfig


@pytest.mark.parametrize("scheduler", [None, "CosineAnnealingLR"])
@pytest.mark.parametrize("enable_amp", [False, True])
def test_optimization_reload(
    scheduler: Literal["CosineAnnealingLR"] | None,
    enable_amp: bool,
):
    """
    Test that when we save and reload the optimizer, we get an identical
    final training result as if we didn't save and reload.
    """
    torch.manual_seed(0)
    optimizer_type: Literal["Adam", "FusedAdam"] = "Adam"
    lr = 0.001
    max_epochs = 10  # final epoch
    checkpoint_epoch = 5  # epoch where we will save and later restart training
    # set up a toy single-layer model with random training data to optimize
    model = nn.Linear(1, 1).to(fme.get_device())
    x = torch.randn(100, 1).to(fme.get_device())
    if scheduler == "CosineAnnealingLR":
        scheduler_config = SchedulerConfig(
            type="CosineAnnealingLR",
            kwargs={"T_max": max_epochs},
        )
    elif scheduler is None:
        scheduler_config = SchedulerConfig()
    else:
        raise NotImplementedError()
    optimization = Optimization(
        parameters=model.parameters(),
        optimizer_type=optimizer_type,
        lr=lr,
        max_epochs=max_epochs,
        scheduler=scheduler_config,
        enable_automatic_mixed_precision=enable_amp,
        kwargs={},
    )
    # train the model
    optimization.set_mode(nn.ModuleList([model]))
    model_intermediate_state = None
    for i in range(max_epochs):
        if i == checkpoint_epoch:
            # save the state
            intermediate_state = yaml.dump(optimization.get_state())
            # save the model weights
            model_intermediate_state = copy.deepcopy(model.state_dict())
        loss = model(x).sum()
        optimization.accumulate_loss(loss)
        optimization.step_weights()
        if scheduler is not None:
            optimization.step_scheduler(loss.item())
    model_first_final_state = copy.deepcopy(model.state_dict())
    # reset the model
    model = nn.Linear(1, 1).to(fme.get_device())
    # reload the model weights
    model.load_state_dict(model_intermediate_state)
    # reload the state
    optimization = Optimization(
        parameters=model.parameters(),
        optimizer_type=optimizer_type,
        lr=lr,
        max_epochs=max_epochs,
        scheduler=scheduler_config,
        enable_automatic_mixed_precision=enable_amp,
        kwargs={},
    )
    optimization.load_state(yaml.load(intermediate_state, Loader=yaml.CLoader))
    # train the model again
    optimization.set_mode(nn.ModuleList([model]))
    for i in range(max_epochs - checkpoint_epoch):
        loss = model(x).sum()
        optimization.accumulate_loss(loss)
        optimization.step_weights()
        if scheduler is not None:
            optimization.step_scheduler(loss.item())
    model_second_final_state = model.state_dict()
    # check that the final weights are the same
    for k in model_first_final_state.keys():
        assert torch.allclose(model_first_final_state[k], model_second_final_state[k])


def test_adam_reload():
    """
    Test that when we save and reload the Adam optimizer, we get an identical
    final training result as if we didn't save and reload.

    This is an external dependency test only.
    """
    torch.manual_seed(0)
    lr = 0.001
    max_epochs = 6  # final epoch
    checkpoint_epoch = 3  # epoch where we will save and later restart training
    # set up a toy single-layer model with random training data to optimize
    model = nn.Linear(1, 1).to(fme.get_device())
    x = torch.randn(100, 1).to(fme.get_device())

    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    model_intermediate_state = None
    # train the model
    for i in range(max_epochs):
        if i == checkpoint_epoch:
            # save the state
            intermediate_state = yaml.dump(optimizer.state_dict())
            # save the model weights
            model_intermediate_state = copy.deepcopy(model.state_dict())
        loss = model(x).sum()
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
    model_first_final_state = copy.deepcopy(model.state_dict())
    # reset the model
    model = nn.Linear(1, 1).to(fme.get_device())
    # reload the model weights
    model.load_state_dict(model_intermediate_state)
    # reload the state
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    optimizer.load_state_dict(yaml.load(intermediate_state, Loader=yaml.CLoader))
    # train the model again
    for i in range(max_epochs - checkpoint_epoch):
        loss = model(x).sum()
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
    model_second_final_state = model.state_dict()
    # check that the final weights are the same
    for k in model_first_final_state.keys():
        assert torch.allclose(model_first_final_state[k], model_second_final_state[k])


def test_null_optimization_accumulates_loss():
    optimizer = NullOptimization()
    optimizer.accumulate_loss(torch.tensor(1.0))
    optimizer.accumulate_loss(torch.tensor(2.0))
    assert optimizer.get_accumulated_loss() == torch.tensor(3.0)
    optimizer.step_weights()
    assert optimizer.get_accumulated_loss() == torch.tensor(0.0)


def test_adam_optimization_accumulates_loss():
    model = nn.Linear(1, 1).to(fme.get_device())
    input = torch.randn(100, 1).to(fme.get_device())
    output = model(input)
    loss = output.sum()
    optimizer = Optimization(
        parameters=model.parameters(),
        optimizer_type="Adam",
        lr=0.001,
        max_epochs=10,
        scheduler=SchedulerConfig(),
        enable_automatic_mixed_precision=False,
        kwargs={},
    )
    optimizer.accumulate_loss(loss)
    optimizer.accumulate_loss(loss)
    assert optimizer.get_accumulated_loss() == loss * 2
    optimizer.step_weights()
    assert optimizer.get_accumulated_loss() == torch.tensor(0.0)


@pytest.mark.parametrize("use_gradient_accumulation", [True, False])
def test_detach_if_using_gradient_accumulation(use_gradient_accumulation: bool):
    model = nn.Linear(1, 1).to(fme.get_device())
    a = torch.randn(10, 1).to(fme.get_device())
    data = {"a": model(a)}
    optimizer = Optimization(
        parameters=model.parameters(),
        optimizer_type="Adam",
        lr=0.001,
        max_epochs=10,
        scheduler=SchedulerConfig(),
        enable_automatic_mixed_precision=False,
        use_gradient_accumulation=use_gradient_accumulation,
        kwargs={},
    )
    data = optimizer.detach_if_using_gradient_accumulation(data)
    assert data["a"].requires_grad == (not use_gradient_accumulation)


def test_change_identical_with_or_without_gradient_accumulation():
    def get_change(use_gradient_accumulation: bool):
        model = nn.Linear(1, 1).to(fme.get_device())
        model.load_state_dict(
            {
                "weight": torch.ones(1, 1).to(fme.get_device()),
                "bias": torch.zeros(1).to(fme.get_device()),
            }
        )
        optimizer = Optimization(
            parameters=model.parameters(),
            optimizer_type="Adam",
            lr=0.001,
            max_epochs=10,
            scheduler=SchedulerConfig(),
            enable_automatic_mixed_precision=False,
            use_gradient_accumulation=use_gradient_accumulation,
            kwargs={},
        )
        a = torch.ones(10, 1).to(fme.get_device())
        loss = model(a).sum()
        optimizer.accumulate_loss(loss)
        loss = model(a).sum()
        optimizer.accumulate_loss(loss)
        optimizer.step_weights()
        return model.state_dict()

    state_with_gradient_accumulation = get_change(True)
    state_without_gradient_accumulation = get_change(False)
    assert state_with_gradient_accumulation == state_without_gradient_accumulation


@pytest.mark.parametrize("after_n_forward_steps", [0, 1, 2])
@pytest.mark.parametrize("step", [0, 1, 2])
def test_checkpoint(after_n_forward_steps: int, step: int):
    config = CheckpointConfig(after_n_forward_steps=after_n_forward_steps)
    checkpoint = config.build(step)
    if step >= after_n_forward_steps:
        assert isinstance(checkpoint, Checkpoint)
    else:
        assert isinstance(checkpoint, NoCheckpoint)



import numpy as np
import pytest
import torch
from torch import nn

from fme.core.weight_ops import CopyWeightsConfig, overwrite_weights


class SimpleLinearModule(torch.nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear = torch.nn.Linear(in_features, out_features)
        self.custom_param = torch.nn.Parameter(torch.randn(out_features))

    def forward(self, x):
        return self.linear(x) + self.custom_param


class NestedModule(torch.nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear1 = SimpleLinearModule(in_features, out_features)
        self.linear2 = SimpleLinearModule(out_features, in_features)
        self.custom_param = torch.nn.Parameter(torch.randn(5, 5))


@pytest.mark.parametrize(
    "from_module, to_module, expected_exception",
    [
        pytest.param(
            SimpleLinearModule(10, 10),
            SimpleLinearModule(10, 10),
            None,
            id="Matching sizes",
        ),
        pytest.param(
            SimpleLinearModule(10, 10),
            SimpleLinearModule(20, 10),
            None,
            id="to_module larger weights",
        ),
        pytest.param(
            SimpleLinearModule(20, 10),
            SimpleLinearModule(10, 10),
            ValueError,
            id="from_module larger weights",
        ),
        pytest.param(
            NestedModule(10, 20),
            NestedModule(10, 20),
            None,
            id="Complex modules matching sizes",
        ),
        pytest.param(
            NestedModule(10, 20),
            SimpleLinearModule(10, 20),
            ValueError,
            id="Nested modules, mismatched structure",
        ),
    ],
)
def test_overwrite_weights(from_module, to_module, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            overwrite_weights(from_module.state_dict(), to_module)
    else:
        overwrite_weights(from_module.state_dict(), to_module)
        for from_param, to_param in zip(
            from_module.parameters(), to_module.parameters()
        ):
            if len(from_param.shape) == 1:
                assert torch.allclose(
                    from_param.data, to_param.data[: from_param.data.size(0)]
                )
            else:
                assert torch.allclose(
                    from_param.data,
                    to_param.data[: from_param.data.size(0), : from_param.data.size(1)],
                )


def test_overwrite_weights_exclude():
    from_module = NestedModule(10, 20)
    to_module = NestedModule(10, 20)
    overwrite_weights(
        from_module.state_dict(), to_module, exclude_parameters=["linear1.*"]
    )
    assert not torch.allclose(
        from_module.linear1.linear.weight, to_module.linear1.linear.weight
    )
    assert not torch.allclose(
        from_module.linear1.custom_param, to_module.linear1.custom_param
    )
    assert torch.allclose(
        from_module.linear2.linear.weight, to_module.linear2.linear.weight
    )
    assert torch.allclose(
        from_module.linear2.custom_param, to_module.linear2.custom_param
    )
    assert torch.allclose(from_module.custom_param, to_module.custom_param)


class NestedModule2(nn.Module):
    def __init__(self):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(3, 3))


class NestedModule1(nn.Module):
    def __init__(self):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(3, 3))
        self.nested = NestedModule2()


@pytest.mark.parametrize(
    "include, exclude, expected_applied, expected_error",
    [
        pytest.param(["*"], [], ["weight", "nested.weight"], None, id="include all"),
        pytest.param([], ["*"], [], None, id="exclude all"),
        pytest.param(["weight"], ["nested.*"], ["weight"], None, id="weight included"),
        pytest.param(["*"], ["nested.*"], [], ValueError, id="nested param in both"),
        pytest.param(["*"], ["weight"], [], ValueError, id="* include with an exclude"),
        pytest.param([], ["weight"], [], ValueError, id="missing weight using exclude"),
        pytest.param(["weight"], [], [], ValueError, id="missing weight using include"),
        pytest.param(
            ["*.weight"], [], [], ValueError, id="mising weight using wildcard include"
        ),
    ],
)
def test_apply_copy_weights_config(
    include: list[str],
    exclude: list[str],
    expected_applied: list[str],
    expected_error: type[Exception] | None,
):
    source_model = NestedModule1()
    dest_model = NestedModule1()
    original_dest_model_state = dest_model.state_dict()

    if expected_error is not None:
        with pytest.raises(expected_error):
            config = CopyWeightsConfig(
                include=include,
                exclude=exclude,
            )
            config.apply([source_model.state_dict()], [dest_model])
    else:
        config = CopyWeightsConfig(
            include=include,
            exclude=exclude,
        )
        config.apply([source_model.state_dict()], [dest_model])

        for name, param in dest_model.named_parameters():
            if name in expected_applied:
                np.testing.assert_array_equal(
                    param.detach().cpu().numpy(),
                    source_model.state_dict()[name].detach().cpu().numpy(),
                )
            else:
                np.testing.assert_array_equal(
                    param.detach().cpu().numpy(),
                    original_dest_model_state[name].detach().cpu().numpy(),
                )



"""Tests for segmented inference entrypoint."""

import dataclasses
import datetime
import os
import pathlib
import tempfile
import unittest.mock

import cftime
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

import fme
from fme.ace.data_loading.inference import ForcingDataLoaderConfig, TimestampList
from fme.ace.inference.data_writer import DataWriterConfig
from fme.ace.inference.inference import (
    InitialConditionConfig,
    main,
    run_segmented_inference,
)
from fme.ace.registry import ModuleSelector
from fme.ace.stepper import SingleModuleStepperConfig
from fme.ace.testing import DimSizes, FV3GFSData
from fme.core.coordinates import (
    DimSize,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
)
from fme.core.dataset.config import XarrayDataConfig
from fme.core.dataset_info import DatasetInfo
from fme.core.logging_utils import LoggingConfig
from fme.core.normalizer import NormalizationConfig

TIMESTEP = datetime.timedelta(hours=6)


class PlusOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def save_stepper(
    path: pathlib.Path,
    in_names: list[str],
    out_names: list[str],
    mean: float,
    std: float,
    data_shape: list[int],
    timestep: datetime.timedelta = TIMESTEP,
):
    all_names = list(set(in_names).union(out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": PlusOne()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={name: mean for name in all_names},
            stds={name: std for name in all_names},
        ),
    )
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(data_shape[-2]), lon=torch.zeros(data_shape[-1])
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7), bk=torch.arange(7)
    )
    stepper = config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coordinate,
            vertical_coordinate=vertical_coordinate,
            timestep=timestep,
        ),
    )
    torch.save({"stepper": stepper.get_state()}, path)


def test_inference_segmented_entrypoint():
    # we use tempfile here instead of pytest tmp_path fixture, because the latter causes
    # issues with checking last modified time of files produced by the test.
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        forward_steps_in_memory = 2
        in_names = ["prog", "forcing_var"]
        out_names = ["prog", "diagnostic_var"]
        stepper_path = tmp_path / "stepper"
        horizontal = [DimSize("lat", 16), DimSize("lon", 32)]

        dim_sizes = DimSizes(
            n_time=18,
            horizontal=horizontal,
            nz_interface=4,
        )
        save_stepper(
            stepper_path,
            in_names=in_names,
            out_names=out_names,
            mean=0.0,
            std=1.0,
            data_shape=dim_sizes.shape_nd,
        )
        data = FV3GFSData(
            path=tmp_path,
            names=["forcing_var"],
            dim_sizes=dim_sizes,
            timestep_days=0.25,
        )
        initial_condition = xr.Dataset(
            {
                "prog": xr.DataArray(
                    np.random.rand(2, 16, 32).astype(np.float32),
                    dims=["sample", "lat", "lon"],
                )
            }
        )

        initial_condition_path = tmp_path / "init_data" / "ic.nc"
        initial_condition_path.parent.mkdir()
        initial_condition["time"] = xr.DataArray(
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
            dims=["sample"],
        )
        initial_condition.to_netcdf(initial_condition_path, mode="w")
        forcing_loader = ForcingDataLoaderConfig(
            dataset=data.inference_data_loader_config.dataset,
            num_data_workers=0,
        )

        run_dir = tmp_path / "segmented_run"
        config = fme.ace.InferenceConfig(
            experiment_dir=str(run_dir),
            n_forward_steps=3,
            forward_steps_in_memory=forward_steps_in_memory,
            checkpoint_path=str(stepper_path),
            logging=LoggingConfig(
                log_to_screen=True, log_to_file=False, log_to_wandb=False
            ),
            initial_condition=InitialConditionConfig(
                path=str(initial_condition_path),
                start_indices=TimestampList(["2000-01-01T06:00:00"]),
            ),
            forcing_loader=forcing_loader,
            data_writer=DataWriterConfig(save_prediction_files=True),
            allow_incompatible_dataset=True,  # stepper checkpoint has arbitrary info  # noqa: E501
        )

        # run one segment of 3 steps
        config_path = str(tmp_path / "config.yaml")
        with open(config_path, "w") as f:
            yaml.dump(dataclasses.asdict(config), f)
        main(config_path, 1)

        # run another segment of 3 steps, and ensure first segment is not being re-run
        filename = os.path.join(
            run_dir, "segment_0000", "autoregressive_predictions.nc"
        )
        before_second_segment_mtime = os.path.getmtime(filename)
        main(config_path, 2)
        after_second_segment_mtime = os.path.getmtime(filename)
        assert before_second_segment_mtime == pytest.approx(after_second_segment_mtime)

        # do a non-segmented run of 6 steps
        config.n_forward_steps = 6
        config.experiment_dir = str(tmp_path / "non_segmented_run")
        config_path = str(tmp_path / "config.yaml")
        with open(config_path, "w") as f:
            yaml.dump(dataclasses.asdict(config), f)
        main(config_path)

        # assert each segment generated output of correct duration
        ds_two_segments_0 = xr.open_dataset(
            run_dir / "segment_0000" / "autoregressive_predictions.nc",
            decode_timedelta=False,
        )
        ds_two_segments_1 = xr.open_dataset(
            run_dir / "segment_0001" / "autoregressive_predictions.nc",
            decode_timedelta=False,
        )
        assert len(ds_two_segments_0.time) == len(ds_two_segments_1.time)

        # Ensure the second half of the 6-step run matches the second segment of the
        # 3-step run. Before comparing, drop init_time and time coordinates, since
        # we don't expect these to match.
        ds_one_segment = xr.open_dataset(
            tmp_path / "non_segmented_run" / "autoregressive_predictions.nc",
            decode_timedelta=False,
        )
        ds_two_segments_1 = ds_two_segments_1.drop_vars(["init_time", "time"])
        ds_one_segment = ds_one_segment.drop_vars(["init_time", "time"])
        xr.testing.assert_equal(
            ds_two_segments_1, ds_one_segment.isel(time=slice(3, None))
        )


def _run_inference_from_config_mock(config: fme.ace.InferenceConfig):
    if not os.path.exists(config.experiment_dir):
        os.makedirs(config.experiment_dir)
    with open(os.path.join(config.experiment_dir, "restart.nc"), "w") as f:
        f.write("mock restart file")
    with open(os.path.join(config.experiment_dir, "wandb_name_env_var"), "w") as f:
        f.write(os.environ.get("WANDB_NAME", ""))


def _get_mock_config(experiment_dir: str) -> fme.ace.InferenceConfig:
    return fme.ace.InferenceConfig(
        experiment_dir=experiment_dir,
        n_forward_steps=3,
        checkpoint_path="mock_checkpoint",
        logging=LoggingConfig(
            log_to_screen=True, log_to_file=False, log_to_wandb=False
        ),
        initial_condition=InitialConditionConfig(path="mock_ic"),
        forcing_loader=ForcingDataLoaderConfig(
            dataset=XarrayDataConfig(data_path="mock_forcing")
        ),
    )


def test_run_segmented_inference(tmp_path, monkeypatch):
    WRITTEN_WANDB_NAME_FILENAME = "wandb_name_env_var"
    mock = unittest.mock.MagicMock(side_effect=_run_inference_from_config_mock)
    config = _get_mock_config(str(tmp_path))

    with unittest.mock.patch(
        "fme.ace.inference.inference.run_inference_from_config", new=mock
    ):
        # run a single segment
        monkeypatch.setenv("WANDB_NAME", "run_name")
        run_segmented_inference(config, 1)
        segment_dir = os.path.join(config.experiment_dir, "segment_0000")
        expected_restart_path = os.path.join(segment_dir, "restart.nc")
        assert os.path.exists(expected_restart_path)
        assert mock.call_count == 1
        with open(os.path.join(segment_dir, WRITTEN_WANDB_NAME_FILENAME)) as f:
            assert f.read() == "run_name-segment_0000"

        # rerun the same segment and ensure run_inference_from_config isn't called again
        run_segmented_inference(config, 1)
        assert os.path.exists(expected_restart_path)
        assert mock.call_count == 1

        # extend to three segments and ensure exactly three run_inference_from_config
        # calls have been made
        monkeypatch.setenv("WANDB_NAME", "run_name")
        run_segmented_inference(config, 3)
        for i in range(3):
            segment_dir = os.path.join(config.experiment_dir, f"segment_{i:04d}")
            expected_restart_path = os.path.join(segment_dir, "restart.nc")
            assert os.path.exists(expected_restart_path)
            with open(os.path.join(segment_dir, WRITTEN_WANDB_NAME_FILENAME)) as f:
                assert f.read() == f"run_name-segment_{i:04d}"
        assert mock.call_count == 3



import dataclasses
from collections.abc import Callable
from copy import copy
from typing import Any, TypeVar

import torch
from torch import nn

from fme.core.dataset_info import DatasetInfo
from fme.core.multi_call import MultiCall, MultiCallConfig, StepMethod
from fme.core.normalizer import StandardNormalizer
from fme.core.ocean import OceanConfig
from fme.core.step.step import StepABC, StepConfigABC, StepSelector
from fme.core.typing_ import TensorDict, TensorMapping


def replace_multi_call(
    selector: StepSelector, multi_call: MultiCallConfig | None, state: dict[str, Any]
) -> tuple[StepSelector, dict[str, Any]]:
    """
    Replace the multi-call configuration in a StepSelector and ensure the
    associated state can be loaded as a multi-call step.

    A value of `None` for `multi_call` will remove the multi-call configuration.

    If the selected type supports it, the multi-call configuration will be
    updated in place. Otherwise, it will be wrapped in the multi_call step
    configuration with the given multi_call config or None.

    Args:
        selector: StepSelector to replace the multi-call configuration of. If
            the StepSelector does not have "multi_call" type, then the step
            will be wrapped in a "multi_call" type StepSelector.
        multi_call: MultiCallConfig for returned StepSelector.
        state: state dictionary associated with the loaded step.

    Returns:
        A StepSelector with the specified MultiCallConfig and the state
        dictionary updated to ensure consistency with that of a serialized
        multi-call step.
    """
    state_copy = state.copy()
    if selector.type == "multi_call":
        wrapped_selector_dict: dict[str, Any] = selector.config["wrapped_step"]
        include_multi_call_in_loss = selector.config.get(
            "include_multi_call_in_loss", True
        )
        new_state = state_copy
    else:
        wrapped_selector_dict = dataclasses.asdict(selector)
        include_multi_call_in_loss = True
        new_state = {"wrapped_step": state_copy}
    if multi_call is None:
        include_multi_call_in_loss = False
    new_selector = StepSelector(
        type="multi_call",
        config={
            "wrapped_step": wrapped_selector_dict,
            "config": dataclasses.asdict(multi_call) if multi_call else None,
            "include_multi_call_in_loss": include_multi_call_in_loss,
        },
    )
    return new_selector, new_state


@StepSelector.register("multi_call")
@dataclasses.dataclass
class MultiCallStepConfig(StepConfigABC):
    """
    Configuration for a multi-call step.

    Parameters:
        wrapped_step: The step to wrap.
        config: The multi-call configuration.
        include_multi_call_in_loss: Whether to include multi-call diagnostics in the
            loss.
    """

    wrapped_step: StepSelector
    config: MultiCallConfig | None = None
    include_multi_call_in_loss: bool = True

    def __post_init__(self):
        if self.config is not None:
            self.config.validate(
                self.wrapped_step.input_names, self.wrapped_step.output_names
            )
        if self.config is None and self.include_multi_call_in_loss:
            raise ValueError("include_multi_call_in_loss is True, but config is None")

    def get_step(
        self,
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None],
    ) -> "MultiCallStep":
        wrapped = self.wrapped_step.get_step(dataset_info, init_weights)
        if self.config is not None:
            self.config.validate(wrapped.input_names, wrapped.output_names)
        return MultiCallStep(
            wrapped_step=wrapped,
            config=self,
        )

    def build(
        self,
        step_method: StepMethod,
    ) -> "MultiCall | None":
        if self.config is None:
            return None
        else:
            return self.config.build(step_method)

    def extend_normalizer_with_multi_call_outputs(
        self, normalizer: StandardNormalizer
    ) -> StandardNormalizer:
        """
        Extend the normalizer by setting multi-call output names to use the same
        normalization as their base counterparts.
        """
        if self.config is None:
            return normalizer
        else:
            return _extend_normalizer_with_multi_call_outputs(self.config, normalizer)

    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        """
        Get the loss normalizer for the multi-call step.

        Normalizer will use statistics from multi-call variables in the stats
        dataset, meaning the normalization for multi-call output versions will be
        different from the normalization for the base variables.

        Args:
            extra_names: Names of additional variables to include in the
                loss normalizer.
            extra_residual_scaled_names: extra_names which use residual scale factors,
                if enabled.
        """
        if self.config is not None:
            if extra_names is None:
                extra_names = []
            else:
                extra_names = list(extra_names)  # avoid mutating input
            if extra_residual_scaled_names is None:
                extra_residual_scaled_names = []
            else:
                extra_residual_scaled_names = list(extra_residual_scaled_names)
            for output_name in self.config.output_names:
                for name in self.config.get_multi_called_names(output_name):
                    extra_names.append(name)
                    if output_name in self.wrapped_step.input_names:
                        extra_residual_scaled_names.append(name)
        return self.wrapped_step.get_loss_normalizer(
            extra_names=extra_names,
            extra_residual_scaled_names=extra_residual_scaled_names,
        )

    @property
    def _multi_call_outputs(self) -> list[str]:
        if self.config is None:
            return []
        return self.config.names

    @property
    def input_names(self) -> list[str]:
        return self.wrapped_step.input_names

    def get_next_step_forcing_names(self) -> list[str]:
        return self.wrapped_step.get_next_step_forcing_names()

    @property
    def output_names(self) -> list[str]:
        return self.wrapped_step.output_names + self._multi_call_outputs

    @property
    def next_step_input_names(self) -> list[str]:
        return self.wrapped_step.next_step_input_names

    @property
    def loss_names(self) -> list[str]:
        if self.include_multi_call_in_loss:
            return self.wrapped_step.loss_names + self._multi_call_outputs
        else:
            return self.wrapped_step.loss_names

    def replace_ocean(self, ocean: OceanConfig | None):
        self.wrapped_step.replace_ocean(ocean)

    def get_ocean(self) -> OceanConfig | None:
        return self.wrapped_step.get_ocean()

    def replace_multi_call(self, multi_call: MultiCallConfig | None):
        self.config = multi_call

    @property
    def n_ic_timesteps(self) -> int:
        return self.wrapped_step.n_ic_timesteps

    def load(self):
        self.wrapped_step.load()


def _extend_normalizer_with_multi_call_outputs(
    config: MultiCallConfig, normalizer: StandardNormalizer
) -> StandardNormalizer:
    means = copy(normalizer.means)
    stds = copy(normalizer.stds)
    for name in config.output_names:
        if name not in means or name not in stds:
            raise ValueError(
                f"Normalizer does not contain {name} present in multi-call output names"
            )
        for multi_call_name in config.get_multi_called_names(name):
            means[multi_call_name] = means[name]
            stds[multi_call_name] = stds[name]
    return StandardNormalizer(
        means=means,
        stds=stds,
        fill_nans_on_normalize=normalizer.fill_nans_on_normalize,
        fill_nans_on_denormalize=normalizer.fill_nans_on_denormalize,
    )


class MultiCallStep(StepABC):
    """
    Step class for a single pytorch module.
    """

    SelfType = TypeVar("SelfType", bound="MultiCallStep")

    TIME_DIM = 1
    CHANNEL_DIM = -3

    def __init__(
        self,
        wrapped_step: StepABC,
        config: MultiCallStepConfig,
    ):
        """
        Args:
            wrapped_step: The step to wrap.
            config: The multi-call step configuration.
        """
        super().__init__()
        self._wrapped_step = wrapped_step
        self._config = config
        self._multi_call = config.build(self._wrapped_step.step)
        self._include_multi_call_in_loss = config.include_multi_call_in_loss

    @property
    def config(self) -> MultiCallStepConfig:
        return self._config

    @property
    def modules(self) -> torch.nn.ModuleList:
        return self._wrapped_step.modules

    @property
    def normalizer(self) -> StandardNormalizer:
        return self._config.extend_normalizer_with_multi_call_outputs(
            self._wrapped_step.normalizer
        )

    @property
    def surface_temperature_name(self) -> str | None:
        return self._wrapped_step.surface_temperature_name

    @property
    def ocean_fraction_name(self) -> str | None:
        return self._wrapped_step.ocean_fraction_name

    def get_regularizer_loss(self) -> torch.Tensor:
        return self._wrapped_step.get_regularizer_loss()

    def step(
        self,
        input: TensorMapping,
        next_step_input_data: TensorMapping,
        wrapper: Callable[[torch.nn.Module], torch.nn.Module] = lambda x: x,
    ) -> TensorDict:
        state = self._wrapped_step.step(
            input,
            next_step_input_data,
            wrapper=wrapper,
        )
        if self._multi_call is not None:
            multi_called_outputs = self._multi_call.step(
                input, next_step_input_data, wrapper=wrapper
            )
            state = {**multi_called_outputs, **state}
        return state

    def get_state(self) -> dict[str, Any]:
        """
        Get the ML model state of the multi-call step.

        Returns:
            The ML model state of the multi-call step.
        """
        return {
            "wrapped_step": self._wrapped_step.get_state(),
        }

    def load_state(self, state: dict[str, Any]):
        """
        Load the ML model state of the multi-call step.

        Does not load the multi-call configuration.

        Args:
            state: The ML model state of the multi-call step.
        """
        self._wrapped_step.load_state(state["wrapped_step"])



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TypeVar

T = TypeVar("T")


def get_all_names(
    *data_varnames: Iterable[T], allowlist: Iterable[T] | None = None
) -> set[T]:
    """
    Returns all variable names from lists of variable names, optionally
    filtering by an allowlist.
    """
    variables: set[T] = set()
    for varnames in data_varnames:
        variables = variables.union(set(varnames))
    if allowlist is None:
        return variables
    else:
        return variables.intersection(set(allowlist))


@dataclass
class DimInfo:
    name: str
    index: int


DIM_INFO_LATLON = [
    DimInfo(name="lat", index=-2),
    DimInfo(name="lon", index=-1),
]

DIM_INFO_HEALPIX = [
    DimInfo(name="face", index=-3),
    DimInfo(name="height", index=-2),
    DimInfo(name="width", index=-1),
]



import os

import torch

from .typing_ import TensorDict, TensorMapping


def using_gpu() -> bool:
    return get_device().type == "cuda"


def get_device() -> torch.device:
    """If CUDA is available, return a CUDA device. Otherwise, return a CPU device
    unless FME_USE_MPS is set, in which case return an MPS device if available.
    """
    if torch.cuda.is_available():
        return torch.device("cuda", torch.cuda.current_device())
    else:
        mps_available = torch.backends.mps.is_available()
        if mps_available and os.environ.get("FME_USE_MPS", "0") == "1":
            return torch.device("mps", 0)
        else:
            return torch.device("cpu")


def move_tensordict_to_device(data: TensorMapping) -> TensorDict:
    device = get_device()
    return {name: value.to(device) for name, value in data.items()}



import dataclasses
from collections.abc import Callable, Mapping

import matplotlib.pyplot as plt
import numpy as np
import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import Image, WandB

from ..plotting import get_cmap_limits, plot_imshow


@dataclasses.dataclass
class _RawData:
    datum: torch.Tensor
    caption: str
    metadata: VariableMetadata
    vmin: float | None = None
    vmax: float | None = None
    cmap: str | None = None

    def get_image(self) -> Image:
        # images are y, x from upper left corner
        # data is time, lat
        # we want lat on y-axis (increasing upward) and time on x-axis
        # so transpose and flip along lat axis
        datum = np.flip(self.datum.transpose(), axis=0)
        fig = plot_imshow(
            datum, vmin=self.vmin, vmax=self.vmax, cmap=self.cmap, flip_lat=False
        )
        wandb = WandB.get_instance()
        wandb_image = wandb.Image(fig, caption=self.caption)
        plt.close(fig)
        return wandb_image


class ZonalMeanAggregator:
    """Images of the zonal-mean state as a function of latitude and time.

    This aggregator keeps track of the generated and target zonal-mean state,
    then generates zonal-mean (Hovmoller) images when logs are retrieved.
    The zonal-mean images are averaged across the sample dimension.
    """

    _captions = {
        "error": (
            "{name} zonal-mean error (generated - target) [{units}], "
            "x-axis is time increasing to right, y-axis is latitude increasing upward"
        ),
        "gen": (
            "{name} zonal-mean generated [{units}], "
            "x-axis is time increasing to right, y-axis is latitude increasing upward"
        ),
    }

    def __init__(
        self,
        zonal_mean: Callable[[torch.Tensor], torch.Tensor],
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        """
        Args:
            zonal_mean: Function that computes the zonal mean of a tensor.
            n_timesteps: Number of timesteps of inference that will be run.
            variable_metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
        """
        self._n_timesteps = n_timesteps
        self._dist = Distributed.get_instance()
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata

        self._target_data: TensorDict | None = None
        self._gen_data: TensorDict | None = None
        self._n_batches = torch.zeros(
            n_timesteps, dtype=torch.int32, device=get_device()
        )[None, :, None]  # sample, time, lat
        self._zonal_mean = zonal_mean

    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int,
    ):
        if self._target_data is None:
            self._target_data = self._initialize_zeros_zonal_mean_from_batch(
                target_data, self._n_timesteps
            )
        if self._gen_data is None:
            self._gen_data = self._initialize_zeros_zonal_mean_from_batch(
                gen_data, self._n_timesteps
            )

        window_steps = next(iter(target_data.values())).shape[1]
        time_slice = slice(i_time_start, i_time_start + window_steps)
        # we can average along longitude without area weighting
        for name, tensor in target_data.items():
            if name in self._target_data:
                self._target_data[name][:, time_slice, :] += self._zonal_mean(tensor)
        for name, tensor in gen_data.items():
            if name in self._gen_data:
                self._gen_data[name][:, time_slice, :] += self._zonal_mean(tensor)
        self._n_batches[:, time_slice, :] += 1

    def _get_data(self) -> dict[str, _RawData]:
        if self._gen_data is None or self._target_data is None:
            raise RuntimeError("No data recorded")
        sample_dim = 0
        data: dict[str, _RawData] = {}
        sorted_names = sorted(list(self._gen_data.keys()))
        for name in sorted_names:
            gen = (
                self._dist.reduce_mean(self._gen_data[name] / self._n_batches)
                .mean(sample_dim)
                .cpu()
                .numpy()
            )
            error = (
                self._dist.reduce_mean(
                    (self._gen_data[name] - self._target_data[name]) / self._n_batches
                )
                .mean(sample_dim)
                .cpu()
                .numpy()
            )

            metadata = self._variable_metadata.get(
                name, VariableMetadata("unknown_units", name)
            )
            vmin, vmax = get_cmap_limits(gen)
            data[f"gen/{name}"] = _RawData(
                datum=gen,
                caption=self._get_caption("gen", name, vmin, vmax),
                # generated data is not considered to have units
                metadata=VariableMetadata(units="", long_name=metadata.long_name),
            )
            vmin, vmax = get_cmap_limits(error, diverging=True)
            data[f"error/{name}"] = _RawData(
                datum=error,
                caption=self._get_caption("error", name, vmin, vmax),
                metadata=metadata,
                vmin=vmin,
                vmax=vmax,
                cmap="RdBu_r",
            )

        return data

    def get_logs(self, label: str) -> dict[str, Image]:
        logs = {}
        data = self._get_data()
        for key, datum in data.items():
            logs[f"{label}/{key}"] = datum.get_image()
        return logs

    def get_dataset(self) -> xr.Dataset:
        data = {
            k.replace("/", "-"): xr.DataArray(
                v.datum, dims=("forecast_step", "lat"), attrs=v.metadata._asdict()
            )
            for k, v in self._get_data().items()
        }

        ret = xr.Dataset(data)
        return ret

    def _get_caption(self, key: str, varname: str, vmin: float, vmax: float) -> str:
        if varname in self._variable_metadata:
            caption_name = self._variable_metadata[varname].long_name
            units = self._variable_metadata[varname].units
        else:
            caption_name, units = varname, "unknown_units"
        caption = self._captions[key].format(name=caption_name, units=units)
        caption += f" vmin={vmin:.4g}, vmax={vmax:.4g}."
        return caption

    @staticmethod
    def _initialize_zeros_zonal_mean_from_batch(
        data: TensorMapping, n_timesteps: int, lat_dim: int = 2
    ) -> TensorDict:
        return {
            name: torch.zeros(
                (tensor.shape[0], n_timesteps, tensor.shape[lat_dim]),
                dtype=tensor.dtype,
                device=tensor.device,
            )
            for name, tensor in data.items()
        }



import datetime
from collections.abc import Mapping
from typing import Any, Literal

import cftime
import matplotlib.pyplot as plt
import numpy as np
import torch
import xarray as xr

from fme.ace.aggregator.plotting import get_cmap_limits, plot_imshow, plot_paneled_data
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import WandB

from .historical_index import INDEX_CALENDAR, NINO34_INDEX

OVERLAP_THRESHOLD = 0.9


def index_data_array(
    index_data: list[tuple[tuple[int, int, int], float]], calendar: str
) -> xr.DataArray:
    """Convert a list of (time, index) tuples to an xarray DataArray.

    Args:
        index_data: List of (time, index) tuples.
        calendar: Calendar for the time coordinate.

    Returns:
        ENSO index data as an xarray DataArray.
    """
    timestamps, index_values = zip(*index_data)
    time_coord = xr.DataArray(
        [cftime.datetime(*timestamp, calendar=calendar) for timestamp in timestamps],
        dims=["time"],
    )
    return xr.DataArray(list(index_values), coords={"time": time_coord}, dims=["time"])


class EnsoCoefficientEvaluatorAggregator:
    _image_captions = {
        "target_gen_coefficient_maps": (
            "{name} target (L) and generated (R) coefficient "
            "with 1940-2020 CMIP6 AMIP Nino 3.4 index [{units} / K]; "
            "metric is not meaningful if a different SST dataset or "
            "interactive ocean is used;"
        ),
        "coefficient_difference_map": (
            "{name} difference in coefficient "
            "with 1940-2020 CMIP6 AMIP Nino 3.4 (generated - target) [{units} / K];"
            "metric is not meaningful if a different SST dataset or "
            "interactive ocean is used;"
        ),
    }

    _enso_index: xr.DataArray = index_data_array(NINO34_INDEX, INDEX_CALENDAR)

    """Compute coefficients of variables regressed against a pre-computed
    scalar ENSO index (i.e., a global-mean time series of ENSO index values).

    We compute the spatially-varying regression coefficients (a map) of each
    predicted variable against the ENSO index, using a simplified covariance-
    over-variance formula for the coefficients that assumes that the index has
    zero mean. For variable i, the coefficients are given by:

        coefficient_i = sum_t(data_it * index_t) / sum_t(index_t ** 2)

    Because the index has zero-mean, this is equivalent to the covariance of the
    data with the index divided by the variance of the index. The computation is
    implemented via running sums over predicted timesteps t, which has the
    advantage of not depending on the number of timesteps in the data, or whether
    data are time-aggregated.

    Args:
        initial_time: Initial time for each sample.
        n_forward_timesteps: Number of timesteps for each sample.
        timestep: Timestep duration.
        gridded_operations: GriddedOperations instance for area-weighted RMSE.
        variable_metadata: Metadata for the variables in the data.
    """

    def __init__(
        self,
        initial_time: xr.DataArray,
        n_forward_timesteps: int,
        timestep: datetime.timedelta,
        gridded_operations: GriddedOperations,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        self._sample_index_series: list[xr.DataArray | None] = get_sample_index_series(
            self.enso_index, initial_time, n_forward_timesteps, timestep
        )
        self._ops = gridded_operations
        if variable_metadata is not None:
            self._variable_metadata: Mapping[str, VariableMetadata] = variable_metadata
        else:
            self._variable_metadata = {}
        n_samples = len(self._sample_index_series)
        self._target_covariances: list[TensorDict] = [{} for _ in range(n_samples)]
        self._gen_covariances: list[TensorDict] = [{} for _ in range(n_samples)]
        self._index_variance: list[torch.Tensor] = [
            torch.tensor(0.0, dtype=torch.float32, device=get_device())
            for _ in range(n_samples)
        ]

    @property
    def enso_index(self) -> xr.DataArray:
        return self._enso_index

    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ):
        """Record running sums of the enso index variance, and of the
        covariance of the target and generated data with the ENSO index (sum
        of squares is used as a proxy for variance/covariance).

        We need to track sums for each sample since the index will be different
        for each time period.
        """
        assert time.sizes["sample"] == len(
            self._sample_index_series
        ), "number of index series must match number of samples"
        for i_sample, sample_index_series in enumerate(self._sample_index_series):
            if sample_index_series is not None:
                sample_index_series_window = sample_index_series.sel(
                    time=time.isel(sample=i_sample)
                )
                sample_index_series_window = torch.tensor(
                    sample_index_series_window.values,
                    device=get_device(),
                    dtype=torch.float32,
                )
                self._index_variance[i_sample] += (sample_index_series_window**2).sum()
                for name, data in target_data.items():
                    if name not in self._target_covariances[i_sample]:
                        self._target_covariances[i_sample][name] = (
                            data_index_covariance(
                                data[i_sample, :], sample_index_series_window
                            )
                        )
                    else:
                        self._target_covariances[i_sample][name] += (
                            data_index_covariance(
                                data[i_sample, :], sample_index_series_window
                            )
                        )
                for name, data in gen_data.items():
                    if name not in self._gen_covariances[i_sample]:
                        self._gen_covariances[i_sample][name] = data_index_covariance(
                            data[i_sample, :], sample_index_series_window
                        )
                    else:
                        self._gen_covariances[i_sample][name] += data_index_covariance(
                            data[i_sample, :], sample_index_series_window
                        )

    def _compute_coefficients(
        self, which: Literal["target", "gen"]
    ) -> list[TensorDict]:
        """Compute the coefficients of the target or generated data regressed
        against the ENSO index for each spatial grid cell for each sample.
        """
        if which == "target":
            covariances = self._target_covariances
        elif which == "gen":
            covariances = self._gen_covariances
        coefficients: list[TensorDict] = [{} for _ in range(len(covariances))]
        for i_sample in range(len(self._index_variance)):
            if self._sample_index_series[i_sample] is not None:
                for name, covariance in covariances[i_sample].items():
                    coefficients[i_sample][name] = (
                        covariance / self._index_variance[i_sample]
                    ).to(device=get_device())
        return coefficients

    def _get_coefficients(self) -> tuple[TensorDict | None, TensorDict | None]:
        dist = Distributed.get_instance()
        target_coefficients = self._compute_coefficients("target")
        gen_coefficients = self._compute_coefficients("gen")
        # average coefficients across samples
        target_coefficients_all, gen_coefficients_all = {}, {}
        target_names = set(
            [
                name
                for target_coefficient in target_coefficients
                for name in target_coefficient.keys()
            ]
        )
        for name in target_names:
            target_coefficients_all[name] = (
                torch.stack(
                    [
                        target_coefficient[name]
                        for target_coefficient in target_coefficients
                        if name in target_coefficient
                    ],
                    dim=0,
                )
                .mean(dim=0)
                .to(device=get_device())
            )
        gen_names = set(
            [
                name
                for gen_coefficient in gen_coefficients
                for name in gen_coefficient.keys()
            ]
        )
        for name in gen_names:
            gen_coefficients_all[name] = (
                torch.stack(
                    [
                        gen_coefficient[name]
                        for gen_coefficient in gen_coefficients
                        if name in gen_coefficient
                    ],
                    dim=0,
                )
                .mean(dim=0)
                .to(device=get_device())
            )
        # average coefficients across processes
        if target_coefficients_all:
            reduced_target_coefficients = reduce_data(dist, target_coefficients_all)
        else:
            reduced_target_coefficients = None
        if gen_coefficients_all:
            reduced_gen_coefficients = reduce_data(dist, gen_coefficients_all)
        else:
            reduced_gen_coefficients = None
        return reduced_target_coefficients, reduced_gen_coefficients

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Any]:
        target_coefficients, gen_coefficients = self._get_coefficients()
        if target_coefficients is None or gen_coefficients is None:
            return {}  # only the root process returns logs
        wandb = WandB.get_instance()
        images, metrics = {}, {}
        for name in gen_coefficients.keys():
            if name in self._variable_metadata:
                caption_name = self._variable_metadata[name].long_name
                caption_units = self._variable_metadata[name].units
            else:
                caption_name = name
                caption_units = "unknown units"
            caption = self._image_captions["target_gen_coefficient_maps"].format(
                name=caption_name, units=caption_units
            )
            panels = [
                [
                    target_coefficients[name].cpu().numpy(),
                    gen_coefficients[name].cpu().numpy(),
                ]
            ]
            coefficient_map = plot_paneled_data(
                data=panels, diverging=True, caption=caption
            )
            images.update({f"coefficient_maps/{name}": coefficient_map})
            rmse = float(
                self._ops.area_weighted_rmse(
                    predicted=gen_coefficients[name],
                    truth=target_coefficients[name],
                    name=name,
                )
                .cpu()
                .numpy()
            )
            metrics.update({f"rmse/{name}": rmse})
            diff = gen_coefficients[name] - target_coefficients[name]
            caption = self._image_captions["coefficient_difference_map"].format(
                name=caption_name, units=caption_units
            )
            vmin, vmax = get_cmap_limits(diff.cpu().numpy(), diverging=True)
            caption += (
                f" vmin={vmin:.4g}, vmax={vmax:.4g}; global-mean RMSE={rmse:.4g}."
            )
            fig = plot_imshow(diff.cpu().numpy(), vmin, vmax, cmap="RdBu_r")
            coefficient_difference_map = wandb.Image(fig, caption=caption)
            plt.close(fig)
            images.update(
                {
                    f"coefficient_difference_map/{name}": coefficient_difference_map,
                }
            )
        logs: dict[str, Any] = {}
        if len(label) > 0:
            label = label + "/"
        logs.update({f"{label}{name}": images[name] for name in images.keys()})
        logs.update({f"{label}{name}": metrics[name] for name in metrics.keys()})
        return logs

    def get_dataset(self) -> xr.Dataset:
        """Get the coefficients as an xarray Dataset."""
        target_coefficients, gen_coefficients = self._get_coefficients()
        if target_coefficients is None or gen_coefficients is None:
            return xr.Dataset()
        target_coefficients_ds = xr.Dataset(
            {
                name: (
                    ["lat", "lon"],
                    target_coefficients[name].cpu().numpy(),
                    self._get_var_attrs(name),
                )
                for name in target_coefficients.keys()
            }
        ).expand_dims({"source": ["target"]})
        gen_coefficients_ds = xr.Dataset(
            {
                name: (["lat", "lon"], gen_coefficients[name].cpu().numpy())
                for name in gen_coefficients.keys()
            }
        ).expand_dims({"source": ["prediction"]})
        return xr.concat([target_coefficients_ds, gen_coefficients_ds], dim="source")

    def _get_var_attrs(self, name: str) -> dict[str, str]:
        if name in self._variable_metadata:
            attrs_name = self._variable_metadata[name].long_name
            attrs_units = self._variable_metadata[name].units
        else:
            attrs_name = name
            attrs_units = "unknown units"
        return {
            "long_name": f"{attrs_name} regression coefficient with Nino3.4 index",
            "units": f"{attrs_units} / K",
        }


def get_sample_index_series(
    index_data: xr.DataArray,
    initial_time: xr.DataArray,
    n_forward_timesteps: int,
    timestep: datetime.timedelta,
    overlap_threshold: float = OVERLAP_THRESHOLD,
) -> list[xr.DataArray | None]:
    """Get a zero-mean index series for each sample, based on the time that
    sample will run for.

    Args:
        index_data: ENSO index data with a time coordinate.
        initial_time: Initial time for each sample.
        n_forward_timesteps: Number of forward timesteps for each sample.
        timestep: Timestep duration.
        overlap_threshold: Required overlap of reference index with inference period.

    Returns:
        List of zero-mean index series for each sample, or None if the sample does
        not overlap sufficiently with the reference index.
    """
    data_calendar = initial_time.dt.calendar
    index_calendar = index_data.time.dt.calendar
    if data_calendar != index_calendar:
        index_data = index_data.convert_calendar(
            calendar=data_calendar, dim="time", use_cftime=True
        )
    sample_index_series: list[xr.DataArray | None] = []
    for initial_time_sample in initial_time:
        duration = n_forward_timesteps * timestep
        end_time = initial_time_sample + duration
        # select index data that overlaps with the inference period, plus a
        # half-timestep buffer since we will later reindex with nearest neighbor
        index_timestep_seconds = (
            index_data.time[1].item() - index_data.time[0].item()
        ).total_seconds()
        half_index_timestep = datetime.timedelta(seconds=index_timestep_seconds / 2)
        sample_index_data_selection = index_data.sel(
            time=slice(
                initial_time_sample - half_index_timestep,
                end_time + half_index_timestep,
            )
        )
        if sample_index_data_selection.sizes["time"] == 0:
            # no overlap
            sample_index_series.append(None)
        else:
            sample_time = xr.date_range(
                start=initial_time_sample.item(),
                end=end_time.item(),
                freq=f"{int(timestep.total_seconds())}s",
                calendar=data_calendar,
                use_cftime=True,
            )
            valid_sample_time = sample_time.where(
                np.logical_and(
                    sample_time >= sample_index_data_selection.time[0],
                    sample_time <= sample_index_data_selection.time[-1],
                ),
            ).dropna()
            if len(valid_sample_time) > len(sample_time) * overlap_threshold:
                reindexed_series = sample_index_data_selection.reindex(
                    time=sample_time, method="nearest"
                )
                reindexed_series_zero_mean = reindexed_series - reindexed_series.mean(
                    "time"
                )
                sample_index_series.append(reindexed_series_zero_mean)
            else:
                # insufficient overlap
                sample_index_series.append(None)
    return sample_index_series


def data_index_covariance(
    data: torch.Tensor, index_values: torch.Tensor, index_dim: int = 0
) -> torch.Tensor:
    """Compute the product of n-dimensional data and a 1-d index,
    where one of the data dimensions aligns with the index. Covariance
    here just means sum of the products of the data and index values.

    Args:
        data: Data tensor with shape that includes index dimension.
        index_values: Index values with shape (n_index,)
        index_dim: Dimension of the index in the data tensor.

    Returns:
        Covariance of data with index
    """
    n_index = index_values.size(0)
    assert data.size(index_dim) == n_index, "aligned dimension must match index length"
    view_dims = [1 if i != index_dim else n_index for i in range(data.dim())]
    index_values_broadcast = index_values.view(*view_dims)
    return (data * index_values_broadcast).sum(dim=index_dim)


def reduce_data(dist: Distributed, rank_tensor_dict: TensorDict) -> TensorDict | None:
    """Reduce tensor dicts across distributed processes by taking the mean.

    Args:
        dist: Distributed instance.
        rank_tensor_dict: Tensor dict to reduce.

    Returns:
        Reduced tensor dict.
    """
    if dist.is_distributed():
        # sort for determinism
        names = sorted(list(rank_tensor_dict.keys()))
        rank_tensor = torch.stack([rank_tensor_dict[name] for name in names], dim=0)
        reduced_tensor = dist.reduce_mean(rank_tensor)
        gathered_tensor_dict = {name: reduced_tensor[i] for i, name in enumerate(names)}
    else:
        gathered_tensor_dict = rank_tensor_dict
    if dist.is_root():
        return gathered_tensor_dict
    else:
        return None



LATENT_HEAT_OF_VAPORIZATION = 2.5e6  # J/kg
LATENT_HEAT_OF_FREEZING = 334000.0  # J/kg
GRAVITY = 9.80665  # m/s^2
# following values are used by SHiELD's slab ocean model, and so we follow suit here.
SPECIFIC_HEAT_OF_WATER = 4000.0  # J/kg/K
DENSITY_OF_WATER = 1000.0  # kg/m^3

SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE = 1004.6  # J/kg/K

RVGAS = 461.5  # J/kg/K
RDGAS = 287.05  # J/kg/K

SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME = SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE - RDGAS

SPECIFIC_HEAT_OF_WATER_CM4 = 3992.0  # J/kg/K
DENSITY_OF_WATER_CM4 = 1035.0  # kg/m^3



import contextlib

import torch

from fme.core import distributed


class MockDistributed:
    def __init__(self, fill_value: float, world_size: int):
        self.world_size = world_size
        self.fill_value = fill_value
        self.reduce_called = False

    def local_batch_size(self, batch_size: int) -> int:
        return batch_size

    def reduce_mean(self, tensor: torch.Tensor) -> torch.Tensor:
        tensor.fill_(self.fill_value)
        self.reduce_called = True
        return tensor

    def reduce_max(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor + 1

    def reduce_sum(self, tensor: torch.Tensor) -> torch.Tensor:
        tensor.fill_(self.fill_value)
        self.reduce_called = True
        return tensor

    def is_root(self) -> bool:
        return True

    def is_distributed(self) -> bool:
        return True

    def gather(self, tensor: torch.Tensor) -> list[torch.Tensor]:
        return [tensor for i in range(self.world_size)]

    def gather_irregular(self, tensor: torch.Tensor) -> list[torch.Tensor] | None:
        """
        Note this uses the actual implementation but mocks the underlying
        distributed calls.
        """
        return distributed.gather_irregular(
            tensor, self.reduce_max, self.gather, is_distributed=self.is_distributed()
        )


@contextlib.contextmanager
def mock_distributed(fill_value: float = 0.0, world_size: int = 1):
    """
    Mock the distributed singleton to return a MockDistributed object.

    This is useful for testing that metrics are reduced across processes.

    It will make it so that when any tensor is reduced, it is filled with
    the given fill_value, which can be checked for in tests.
    """
    original = distributed.singleton
    distributed.singleton = MockDistributed(
        fill_value=fill_value, world_size=world_size
    )  # type: ignore
    try:
        yield distributed.singleton
    finally:
        distributed.singleton = original



import argparse
import itertools
import logging
import os
import tempfile
from dataclasses import dataclass
from typing import Generator, List, Optional, Sequence, Tuple
from urllib.parse import urlparse

import apache_beam as beam
import fsspec
import pandas as pd
import xarray as xr
import xarray_beam as xbeam
import zarr
from apache_beam.pipeline import PipelineOptions

GCS_ROOT = (
    "gs://vcm-ml-raw-flexible-retention/2024-03-11-era5-025deg-2D-variables-from-NCAR"
)

TIME_FORMAT = "%Y%m%d%H"

AN_SFC = "e5.oper.an.sfc"
FC_SFC_MEANFLUX = "e5.oper.fc.sfc.meanflux"
INVARIANT = "e5.oper.invariant"

TARGET_CHUNKS = {
    FC_SFC_MEANFLUX: {
        "forecast_initial_time": 1,
        "forecast_hour": 12,
        "latitude": 721,
        "longitude": 1440,
    },
    AN_SFC: {"time": 12, "latitude": 721, "longitude": 1440},
    INVARIANT: {"time": 1, "latitude": 721, "longitude": 1440},
}

TARGET_SPLIT_CHUNKS = {
    FC_SFC_MEANFLUX: {
        "forecast_initial_time": 1,
        "forecast_hour": 1,
        "latitude": 721,
        "longitude": 1440,
    },
    AN_SFC: {"time": 1, "latitude": 721, "longitude": 1440},
    INVARIANT: {"time": 1, "latitude": 721, "longitude": 1440},
}

VARIABLES = {
    FC_SFC_MEANFLUX: [
        "235_040_mtnlwrf",
        "235_035_msdwswrf",
        "235_037_msnswrf",
        "235_036_msdwlwrf",
        "235_038_msnlwrf",
        "235_033_msshf",
        "235_034_mslhf",
        "235_043_mer",
        "235_055_mtpr",
        "235_031_msr",
        "235_054_mvimd",
        "235_053_mtdwswrf",
        "235_039_mtnswrf",
    ],
    AN_SFC: [
        "128_031_ci",
        "128_039_swvl1",
        "128_040_swvl2",
        "128_041_swvl3",
        "128_042_swvl4",
    ],
    INVARIANT: ["128_172_lsm", "128_129_z"],
}

# Throughout this script, "initial_time" refers to the global initial time of
# the dataset, while "start_time" refers to the first time in an individual
# file.
INITIAL_TIME = {
    FC_SFC_MEANFLUX: "1940-01-01T06",
    AN_SFC: "1940-01",
    INVARIANT: "1979-01-01",
}

PERIODS = {
    FC_SFC_MEANFLUX: 2 * 12 * 84,  # two files per month, 84 years
    AN_SFC: 12 * 84,  # one file per month, 84 years
    INVARIANT: 1,
}

INCLUSIVE = {
    FC_SFC_MEANFLUX: "left",
    AN_SFC: "both",
    INVARIANT: "both",  # Just must not be "neither"
}

TIME_DIM = {
    FC_SFC_MEANFLUX: "forecast_initial_time",
    AN_SFC: "time",
    INVARIANT: "time",
}

FREQUENCY = {
    FC_SFC_MEANFLUX: "12h",
    AN_SFC: "1h",
    INVARIANT: None,
}


def get_time_bounds(
    category: str,
    periods: int,
    initial_time: str,
) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:
    if category == FC_SFC_MEANFLUX:
        times = pd.date_range(initial_time, freq="SMS-16", periods=periods + 1)
        return list(zip(times[:-1], times[1:]))
    elif category == AN_SFC:
        start_times = pd.date_range(initial_time, freq="MS", periods=periods)
        end_times = pd.date_range(
            initial_time, periods=periods, freq="ME"
        ) + pd.Timedelta("23h")
        return list(zip(start_times, end_times))
    elif category == INVARIANT:
        return [(pd.Timestamp(initial_time), pd.Timestamp(initial_time))]
    else:
        raise ValueError(f"Unknown category {category}")


def get_path(
    variable: str, start_time: pd.Timestamp, end_time: pd.Timestamp, category: str
) -> str:
    start_label = start_time.strftime(TIME_FORMAT)
    end_label = end_time.strftime(TIME_FORMAT)
    return (
        f"{GCS_ROOT}/{category}/{variable}/"
        f"{category}.{variable}.ll025sc.{start_label}_{end_label}.nc"
    )


def get_file_metadata(
    time_bounds: Sequence[Tuple[pd.Timestamp, pd.Timestamp]],
    variables: Sequence[str],
    category: str,
) -> List[Tuple[pd.Timestamp, pd.Timestamp, str, str, str]]:
    items = []
    for (start_time, end_time), variable in itertools.product(time_bounds, variables):
        path = get_path(variable, start_time, end_time, category)
        item = (start_time, end_time, variable, category, path)
        items.append(item)
    return items


def validate_time(
    ds: xr.Dataset,
    time_dim: str,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    category: str,
) -> None:
    freq = FREQUENCY[category]
    inclusive = INCLUSIVE[category]
    expected = pd.date_range(start_time, end_time, freq=freq, inclusive=inclusive)
    message = "time coordinate of dataset does not match expected time coordinate"
    assert ds.indexes[time_dim].equals(expected), message


def _get_fs(path):
    protocol = urlparse(path).scheme
    if not protocol:
        protocol = "file"

    fs = fsspec.filesystem(protocol)

    return fs


def _get_key_and_ds(start_time, end_time, variable, category, path):
    """Adjust ds and return key, ds tuple"""

    initial_time = INITIAL_TIME[category]
    time_dim = TIME_DIM[category]

    # Open files as in pangeo-forge and as recommended by xarray-beam (i.e.
    # with the default chunks=None to prevent using dask). This makes Dataflow
    # jobs much less expensive.
    ds = xr.open_dataset(path, engine="h5netcdf", cache=False)

    # Drop utc_date variable, which is present in every Dataset, but is
    # redundant with the time coordinate.
    ds = ds.drop_vars(["utc_date"])

    # Check to make sure the times in the file match what is expected, i.e. a
    # time series with a regular frequency starting at the start time written
    # in the filename and ending either on the end time written in the filename
    # or the time interval before depending on the value of inclusive. This is
    # not strictly necessary, but is a sanity check. If it holds for each file,
    # then our calculation of the offset for each file and the total length of
    # the dataset will be correct.
    validate_time(ds, time_dim, start_time, end_time, category)

    # Infer the offset based on the frequency of the files and the initial
    # date. If the frequency is None, since the time dimension only contains
    # one element, set the offset to 0.
    freq = FREQUENCY[category]
    if freq is None:
        offset = 0
    else:
        offset = (start_time - pd.Timestamp(initial_time)) // pd.Timedelta(freq)

    key = xbeam.Key({time_dim: offset}, vars={variable})
    return key, ds


def create_record(
    item: Tuple[pd.Timestamp, pd.Timestamp, str, str, str], local=True
) -> Tuple[xbeam.Key, xr.Dataset]:
    start_time, end_time, variable, category, path = item

    file = fsspec.open(path, "rb").open()
    key, ds = _get_key_and_ds(start_time, end_time, variable, category, file)
    return key, ds


def create_record_local(
    item: Tuple[pd.Timestamp, pd.Timestamp, str, str, str],
) -> Generator[Tuple[xbeam.Key, xr.Dataset], None, None]:
    start_time, end_time, variable, category, path = item
    with tempfile.TemporaryDirectory() as tmpdir:
        fs = _get_fs(path)
        file = os.path.join(tmpdir, os.path.basename(path))
        fs.get(path, file)
        key, ds = _get_key_and_ds(start_time, end_time, variable, category, file)
        yield key, ds


def get_full_time_coordinate(
    time_bounds: List[Tuple[pd.Timestamp, pd.Timestamp]],
    category: str,
    freq: Optional[str],
) -> pd.DatetimeIndex:
    initial_time, _ = time_bounds[0]
    _, final_time = time_bounds[-1]
    inclusive = INCLUSIVE[category]
    return pd.date_range(initial_time, final_time, inclusive=inclusive, freq=freq)


def get_template(
    time_bounds: List[Tuple[pd.Timestamp, pd.Timestamp]],
    variables: Sequence[str],
    category: str,
) -> xr.Dataset:
    # To construct a template we grab the first file for each variable in the
    # category.
    template_files = get_file_metadata(time_bounds[:1], variables, category)
    datasets = []
    for template_file in template_files:
        _, ds = create_record(template_file)
        datasets.append(ds)
    merged = xr.merge(datasets)

    freq = FREQUENCY[category]
    time = get_full_time_coordinate(time_bounds, category, freq)

    time_dim = TIME_DIM[category]
    merged = merged.isel({time_dim: 0}).drop_vars([time_dim])
    chunks = TARGET_CHUNKS[category]
    return xbeam.make_template(merged).expand_dims({time_dim: time}).chunk(chunks)


def test_ds(store: zarr.storage.FSStore) -> zarr.storage.FSStore:
    ds = xr.open_dataset(store, engine="zarr", chunks={})
    print(ds)
    return store


def _get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "output_path", type=str, help="Output path for outputting zarr store"
    )
    parser.add_argument(
        "category",
        help=f"One of {FC_SFC_MEANFLUX!r}, {AN_SFC!r}, {INVARIANT!r}",
    )
    return parser


@dataclass
class OpenNetCDFandSplit(beam.PTransform):
    """Load a single netcdf locally and split into chunks"""

    category: str

    def _load_and_split(self, item):
        # creating record as generator ensures context with downloaded file
        # stays intact it is completely processed in the split loop
        record_generator = create_record_local(item)
        for key, ds in record_generator:
            split_gen = xbeam.split_chunks(
                key, ds, target_chunks=TARGET_SPLIT_CHUNKS[self.category]
            )
            for key, ds in split_gen:
                yield key, ds.load()

    def expand(self, pcoll):
        return pcoll | beam.FlatMap(self._load_and_split)


def main():
    parser = _get_parser()
    args, pipeline_args = parser.parse_known_args()
    category = args.category
    store = os.path.join(args.output_path, f"{category}.zarr")

    time_bounds = get_time_bounds(category, PERIODS[category], INITIAL_TIME[category])
    file_metadata = get_file_metadata(time_bounds, VARIABLES[category], category)
    template = get_template(time_bounds, VARIABLES[category], category)

    logging.basicConfig(level=logging.DEBUG)

    recipe = (
        beam.Create(file_metadata)
        | OpenNetCDFandSplit(category)
        | xbeam.ConsolidateChunks(TARGET_CHUNKS[category])
        | xbeam.ChunksToZarr(store, template, TARGET_CHUNKS[category])
    )

    with beam.Pipeline(options=PipelineOptions(pipeline_args)) as p:
        p | recipe

    test_ds(store)


if __name__ == "__main__":
    main()



import torch


def get_crps(
    gen: torch.Tensor, target: torch.Tensor, alpha: float = 1.0
) -> torch.Tensor:
    """
    Compute the CRPS loss for a single variable at a single timestep.

    Supports almost-fair modification to CRPS from
    https://arxiv.org/html/2412.15832v1, which claims to be helpful in
    avoiding numerical issues with fair CRPS.

    Args:
        gen: The generated ensemble members, of shape [n_batch, n_ensemble, ...].
        target: The target, of shape [n_batch, 1, ...].
        alpha: The alpha value for the CRPS loss. Corresponds to the alpha value
            for "almost fair" CRPS from https://arxiv.org/html/2412.15832v1. Default
            behavior uses fair CRPS (alpha=1.0).

    Returns:
        The CRPS loss.
    """
    if gen.shape[1] != 2:
        raise NotImplementedError(
            "CRPS loss is written here specifically for 2 ensemble members, "
            f"got {gen.shape[1]} ensemble members. "
            "Update this function (and its tests) to support more."
        )
    # CRPS is `E[|X - y|] - 1/2 E[|X - X'|]`
    # below we compute the first term as the average of two ensemble members
    # meaning the 0.5 factor can be pulled out
    alpha = 0.95
    epsilon = (1 - alpha) / 2
    target_term = torch.abs(gen - target).mean(axis=1)
    internal_term = -0.5 * torch.abs(gen[:, 0, ...] - gen[:, 1, ...])
    return target_term + (1 - epsilon) * internal_term


def get_energy_score(
    gen: torch.Tensor,
    target: torch.Tensor,
) -> torch.Tensor:
    """
    Compute the energy score for a single complex-valued variable at a single
    timestep.

    The energy score is defined as

    .. math::

        E[||X - y||^{beta}] - 1/2 E[||X - X'||^{beta}]

    where :math:`X` is the ensemble, :math:`y` is the target, and :math:`||.||`
    is the complex modulus. It is a proper scoring rule for beta in (0, 2). Here
    we use beta=1. See Gneiting and Raftery (2007) [1]_ Section 4.3 for more details.

    Args:
        target: The target tensor without a sample dimension
        prediction: The prediction tensor with a sample dimension
        sample_dim: The dimension of `prediction` corresponding to sample.

    .. [1] https://sites.stat.washington.edu/people/raftery/Research/PDF/Gneiting2007jasa.pdf

    Args:
        gen: The complex-valued generated ensemble members, of shape
            [n_batch, n_ensemble, ...].
        target: The complex-valued target, of shape [n_batch, 1, ...].

    Returns:
        The energy score.
    """
    if gen.shape[1] != 2:
        raise NotImplementedError(
            "Energy score is written here specifically for 2 ensemble members, "
            f"got {gen.shape[1]} ensemble members. "
            "Update this function (and its tests) to support more."
        )
    # CRPS is `E[|X - y|] - 1/2 E[|X - X'|]`
    # below we compute the first term as the average of two ensemble members
    # meaning the 0.5 factor can be pulled out
    target_term = torch.abs(gen - target).mean(axis=1)
    internal_term = -0.5 * torch.abs(gen[:, 0, ...] - gen[:, 1, ...])
    return target_term + internal_term



import datetime
from typing import NamedTuple

import cftime
import numpy as np
import pytest
import torch
import xarray as xr
from netCDF4 import Dataset
from xarray.coding.times import CFDatetimeCoder

from fme.ace.data_loading.batch_data import PairedData
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.main import (
    DataWriter,
    DataWriterConfig,
    PairedDataWriter,
)
from fme.ace.inference.data_writer.raw import get_batch_lead_time_microseconds
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.core.device import get_device

CALENDAR_CFTIME = {
    "julian": cftime.DatetimeJulian,
    "proleptic_gregorian": cftime.DatetimeProlepticGregorian,
    "noleap": cftime.DatetimeNoLeap,
}

SECONDS_PER_HOUR = 3600
MICROSECONDS_PER_SECOND = 1_000_000
TIMESTEP = datetime.timedelta(hours=6)


def test_data_writer_config_save_names():
    variable_names = ["temp", "humidity"]
    kwargs = dict(
        save_prediction_files=False,
        save_monthly_files=False,
        save_histogram_files=False,
    )
    for save_writer in [
        "save_prediction_files",
        "save_monthly_files",
        "save_histogram_files",
    ]:
        kwargs_copy = kwargs.copy()
        kwargs_copy.update({save_writer: True})
        DataWriterConfig(names=variable_names, **kwargs_copy)  # type: ignore


class TestDataWriter:
    class VariableMetadata(NamedTuple):
        units: str
        long_name: str

    @pytest.fixture(params=["julian", "proleptic_gregorian", "noleap"])
    def calendar(self, request):
        """
        These are the calendars for the datasets we tend to use: 'julian'
        for FV3GFS, 'noleap' for E3SM, and 'proleptic_gregorian' for generic
        datetimes in testing.

        Check that output datasets written with each calendar for their
        time coordinate re-load correctly.
        """
        return request.param

    def get_batch_time(
        self, start_time, end_time, freq, n_initial_conditions, calendar="julian"
    ):
        datetime_class = CALENDAR_CFTIME[calendar]
        start_time = datetime_class(*start_time)
        end_time = datetime_class(*end_time)
        batch_time = xr.DataArray(
            xr.date_range(
                start_time,
                end_time,
                freq=freq,
                calendar=calendar,
                use_cftime=True,
            ).values,
            dims="time",
        )
        return xr.concat(
            [batch_time for _ in range(n_initial_conditions)], dim="sample"
        )

    @pytest.fixture
    def sample_metadata(self):
        return {
            "temp": self.VariableMetadata(units="K", long_name="Temperature"),
            "humidity": self.VariableMetadata(units="%", long_name="Relative Humidity"),
        }

    @pytest.fixture
    def sample_target_data(self, request):
        shape = request.param
        data = {
            # sample, time, *horizontal_dims
            "temp": torch.rand(shape),
            "humidity": torch.rand(shape),  # input-only variable
            "pressure": torch.rand(
                shape
            ),  # Extra variable for which there's no metadata
            "precipitation": torch.rand(shape),  # optionally saved
        }
        return data

    @pytest.fixture
    def sample_prediction_data(self, request):
        shape = request.param
        data = {
            # sample, time, lat, lon
            "temp": torch.rand(shape),
            "pressure": torch.rand(
                shape
            ),  # Extra variable for which there's no metadata
            "precipitation": torch.rand(shape),  # optionally saved
        }
        return data

    @pytest.fixture
    def coords(self, request):
        return request.param

    @pytest.mark.parametrize(
        "sample_target_data, sample_prediction_data, coords",
        [
            pytest.param(
                (2, 3, 4, 5),
                (2, 3, 4, 5),
                {"lat": np.arange(4), "lon": np.arange(5)},
                id="LatLon",
            ),
            pytest.param(
                (2, 3, 6, 4, 5),
                (2, 3, 6, 4, 5),
                {"face": np.arange(6), "height": np.arange(4), "width": np.arange(5)},
                id="HEALPix",
            ),
        ],
        indirect=True,
    )
    def test_append_batch(
        self,
        sample_metadata,
        sample_target_data,
        sample_prediction_data,
        tmp_path,
        calendar,
        coords,
    ):
        n_initial_conditions = 2
        n_timesteps = 6
        writer = PairedDataWriter(
            str(tmp_path),
            n_initial_conditions=n_initial_conditions,
            n_timesteps=n_timesteps,
            timestep=TIMESTEP,
            variable_metadata=sample_metadata,
            coords=coords,
            enable_prediction_netcdfs=True,
            enable_video_netcdfs=False,
            enable_monthly_netcdfs=True,
            enable_histogram_netcdfs=True,
            save_names=None,
            dataset_metadata=DatasetMetadata(source={"inference_version": "1.0"}),
        )
        start_time = (2020, 1, 1, 0, 0, 0)
        end_time = (2020, 1, 1, 12, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time,
            end_time=end_time,
            freq="6h",
            n_initial_conditions=n_initial_conditions,
            calendar=calendar,
        )
        writer.append_batch(
            batch=PairedData(
                prediction=sample_prediction_data,
                reference=sample_target_data,
                time=batch_time,
            ),
        )
        start_time_2 = (2020, 1, 1, 18, 0, 0)
        end_time_2 = (2020, 1, 2, 6, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time_2,
            end_time=end_time_2,
            freq="6h",
            n_initial_conditions=n_initial_conditions,
            calendar=calendar,
        )
        writer.append_batch(
            batch=PairedData(
                prediction=sample_prediction_data,
                reference=sample_target_data,
                time=batch_time,
            ),
        )
        writer.flush()

        # Open the file and check the data
        dataset = Dataset(tmp_path / "autoregressive_predictions.nc", "r")
        assert dataset["time"].units == "microseconds"
        assert dataset["init_time"].units == "microseconds since 1970-01-01 00:00:00"
        assert dataset["init_time"].calendar == calendar
        assert "source.inference_version" in dataset.ncattrs()
        assert dataset.getncattr("source.inference_version") == "1.0"
        assert dataset.getncattr("title") == "ACE autoregressive predictions data file"
        horizontal_shape = (4, 5) if "lat" in coords else (6, 4, 5)
        for var_name in set(sample_prediction_data.keys()):
            var_data = dataset.variables[var_name][:]
            assert var_data.shape == (
                n_initial_conditions,
                n_timesteps,
                *horizontal_shape,
            )
            assert not np.isnan(var_data).any(), "unexpected NaNs in prediction data"
            if var_name in sample_metadata:
                assert (
                    dataset.variables[var_name].units == sample_metadata[var_name].units
                )
                assert (
                    dataset.variables[var_name].long_name
                    == sample_metadata[var_name].long_name
                )
            else:
                assert not hasattr(dataset.variables[var_name], "units")
                assert not hasattr(dataset.variables[var_name], "long_name")

        dataset.close()

        # Open the target output file and do smaller set of checks
        dataset = Dataset(tmp_path / "autoregressive_target.nc", "r")
        coord_names = {"time", "init_time", "valid_time", *set(coords)}
        assert set(dataset.variables) == set(sample_target_data) | coord_names
        assert "source.inference_version" in dataset.ncattrs()
        assert dataset.getncattr("source.inference_version") == "1.0"
        assert dataset.getncattr("title") == "ACE autoregressive target data file"

        # Open the file again with xarray and check the time coordinates,
        # since the values opened this way depend on calendar/units
        with xr.open_dataset(
            tmp_path / "autoregressive_predictions.nc",
            decode_times=CFDatetimeCoder(use_cftime=True),
            decode_timedelta=False,  # prefer handling lead times in ints not timedelta
        ) as ds:
            expected_lead_times = xr.DataArray(
                [
                    MICROSECONDS_PER_SECOND * SECONDS_PER_HOUR * i
                    for i in np.arange(0, 31, 6)
                ],
                dims="time",
            ).assign_coords(
                {
                    "time": [
                        MICROSECONDS_PER_SECOND * SECONDS_PER_HOUR * i
                        for i in np.arange(0, 31, 6)
                    ]
                }
            )
            xr.testing.assert_equal(ds["time"], expected_lead_times)
            expected_init_times = xr.DataArray(
                [
                    CALENDAR_CFTIME[calendar](*start_time)
                    for _ in range(n_initial_conditions)
                ],
                dims=["sample"],
            )
            expected_init_times = expected_init_times.assign_coords(
                {"init_time": expected_init_times}
            )
            xr.testing.assert_equal(ds["init_time"], expected_init_times)
            for var_name in set(sample_prediction_data.keys()):
                assert "valid_time" in ds[var_name].coords
                assert "init_time" in ds[var_name].coords
        for source in ["target", "prediction"]:
            histograms = xr.open_dataset(
                tmp_path / f"histograms_{source}.nc", decode_timedelta=False
            )
            actual_var_names = sorted([str(k) for k in histograms.keys()])
            assert histograms.data_vars["temp"].attrs["units"] == "count"
            assert "temp_bin_edges" in actual_var_names
            assert histograms.data_vars["temp_bin_edges"].attrs["units"] == "K"
            counts_per_timestep = histograms["temp"].sum(dim=["bin"])
            same_count_each_timestep = np.all(
                counts_per_timestep.values == counts_per_timestep.values[0]
            )
            assert same_count_each_timestep

        with xr.open_dataset(
            tmp_path / "monthly_mean_predictions.nc", decode_timedelta=False
        ) as ds:
            assert ds.counts.sum() == n_initial_conditions * n_timesteps
            assert np.sum(np.isnan(ds["precipitation"])) == 0
            assert np.sum(np.isnan(ds["temp"])) == 0
            assert np.sum(np.isnan(ds["pressure"])) == 0
            assert np.all(ds.init_time.dt.year.values > 0)
            assert np.all(ds.init_time.dt.year.values >= 0)
            assert np.all(ds.valid_time.dt.month.values >= 0)
            assert ds.attrs["title"] == "ACE monthly predictions data file"
            assert ds.attrs["source.inference_version"] == "1.0"

    @pytest.mark.parametrize(
        "sample_target_data, sample_prediction_data",
        [pytest.param((2, 3, 4, 5), (2, 3, 4, 5), id="LatLon")],
        indirect=True,
    )
    @pytest.mark.parametrize(
        ["save_names"],
        [
            pytest.param(None, id="None"),
            pytest.param(["temp", "humidity", "pressure"], id="subset"),
        ],
    )
    def test_append_batch_save_names(
        self,
        sample_metadata,
        sample_target_data,
        sample_prediction_data,
        tmp_path,
        save_names,
    ):
        n_samples = 2
        writer = PairedDataWriter(
            str(tmp_path),
            n_initial_conditions=n_samples,
            n_timesteps=4,  # unused
            timestep=TIMESTEP,
            variable_metadata=sample_metadata,
            coords={"lat": np.arange(4), "lon": np.arange(5)},
            enable_prediction_netcdfs=True,
            enable_video_netcdfs=False,
            enable_monthly_netcdfs=True,
            save_names=save_names,
            enable_histogram_netcdfs=True,
            dataset_metadata=DatasetMetadata(),
        )
        start_time = (2020, 1, 1, 0, 0, 0)
        end_time = (2020, 1, 1, 12, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time,
            end_time=end_time,
            freq="6h",
            n_initial_conditions=n_samples,
        )
        writer.append_batch(
            batch=PairedData(
                prediction=sample_prediction_data,
                reference=sample_target_data,
                time=batch_time,
            ),
        )
        writer.flush()
        dataset = Dataset(tmp_path / "autoregressive_predictions.nc", "r")
        expected_variables = (
            set(save_names).intersection(sample_prediction_data)
            if save_names is not None
            else set(sample_prediction_data.keys())
        )
        assert set(dataset.variables.keys()) == expected_variables.union(
            {"init_time", "time", "lat", "lon", "valid_time"}
        )
        expected_prediction_variables = set(sample_prediction_data.keys())
        if save_names is not None:
            expected_prediction_variables = expected_prediction_variables.intersection(
                save_names
            )
        dataset = Dataset(tmp_path / "monthly_mean_predictions.nc", "r")
        expected_variables = (
            set(save_names)
            if save_names is not None
            else set(sample_target_data.keys())
        )
        assert set(dataset.variables.keys()) == expected_prediction_variables.union(
            {
                "init_time",
                "time",
                "valid_time",
                "counts",
                "lat",
                "lon",
            }
        )

    @pytest.mark.parametrize(
        "sample_target_data, sample_prediction_data",
        [pytest.param((2, 3, 4, 5), (2, 3, 4, 5), id="LatLon")],
        indirect=True,
    )
    def test_append_batch_data_time_mismatch(
        self, sample_metadata, sample_target_data, sample_prediction_data, tmp_path
    ):
        n_samples = 2
        writer = PairedDataWriter(
            str(tmp_path),
            n_initial_conditions=n_samples,
            n_timesteps=3,
            timestep=TIMESTEP,
            variable_metadata=sample_metadata,
            coords={"lat": np.arange(4), "lon": np.arange(5)},
            enable_prediction_netcdfs=True,
            enable_video_netcdfs=False,
            enable_monthly_netcdfs=True,
            save_names=None,
            enable_histogram_netcdfs=True,
            dataset_metadata=DatasetMetadata(),
        )
        start_time = (2020, 1, 1, 0, 0, 0)
        end_time = (2020, 1, 1, 12, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time,
            end_time=end_time,
            freq="6h",
            n_initial_conditions=n_samples + 1,
        )
        with pytest.raises(ValueError):
            writer.append_batch(
                batch=PairedData(
                    prediction=sample_prediction_data,
                    reference=sample_target_data,
                    time=batch_time,
                ),
            )

    def test_prediction_only_append_batch(self, sample_metadata, tmp_path, calendar):
        n_samples = 2
        n_timesteps = 8
        coarsen_factor = 2
        device = get_device()
        prediction_data = {
            "temp": torch.rand(
                (n_samples, n_timesteps // coarsen_factor, 4, 5), device=device
            ),
            "pressure": torch.rand(
                (n_samples, n_timesteps // coarsen_factor, 4, 5), device=device
            ),
        }
        reference_data = {
            "insolation": torch.rand(
                (n_samples, n_timesteps // coarsen_factor, 4, 5), device=device
            ),
        }
        writer = DataWriter(
            str(tmp_path),
            n_initial_conditions=n_samples,
            n_timesteps=n_timesteps,
            variable_metadata=sample_metadata,
            coords={"lat": np.arange(4), "lon": np.arange(5)},
            timestep=TIMESTEP,
            enable_prediction_netcdfs=True,
            enable_monthly_netcdfs=True,
            save_names=None,
            time_coarsen=TimeCoarsenConfig(coarsen_factor),
            dataset_metadata=DatasetMetadata(source={"inference_version": "1.0"}),
        )
        start_time = (2020, 1, 1, 0, 0, 0)
        end_time = (2020, 1, 1, 18, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time,
            end_time=end_time,
            freq="6h",
            n_initial_conditions=n_samples,
            calendar=calendar,
        )
        writer.append_batch(
            batch=PairedData(
                prediction=prediction_data,
                reference=reference_data,
                time=batch_time,
            ),
        )
        start_time_2 = (2020, 1, 2, 0, 0, 0)
        end_time_2 = (2020, 1, 2, 18, 0, 0)
        batch_time = self.get_batch_time(
            start_time=start_time_2,
            end_time=end_time_2,
            freq="6h",
            n_initial_conditions=n_samples,
            calendar=calendar,
        )
        writer.append_batch(
            batch=PairedData(
                prediction=prediction_data,
                reference=reference_data,
                time=batch_time,
            ),
        )
        writer.flush()

        with xr.open_dataset(
            tmp_path / "autoregressive_predictions.nc",
            decode_timedelta=False,
        ) as ds:
            assert "temp" in ds
            expected_shape = (n_samples, n_timesteps // coarsen_factor, 4, 5)
            assert ds.temp.shape == expected_shape
            assert ds.valid_time.shape == expected_shape[:2]
            assert set(ds.data_vars) == {"temp", "pressure", "insolation"}
            assert ds.attrs["title"] == "ACE autoregressive predictions data file"
            assert ds.attrs["source.inference_version"] == "1.0"

        with xr.open_dataset(
            tmp_path / "monthly_mean_predictions.nc", decode_timedelta=False
        ) as ds:
            assert ds.counts.sum() == n_samples * n_timesteps
            assert np.sum(np.isnan(ds["temp"])) == 0
            assert np.sum(np.isnan(ds["pressure"])) == 0
            assert np.all(ds.init_time.dt.year.values > 0)
            assert np.all(ds.init_time.dt.year.values >= 0)
            assert np.all(ds.valid_time.dt.month.values >= 0)
            assert set(ds.data_vars) == {"temp", "pressure", "insolation"}
            assert set(ds.coords) == {
                "init_time",
                "valid_time",
                "counts",
                "lat",
                "lon",
                "time",
            }
            assert ds.attrs["title"] == "ACE monthly predictions data file"
            assert ds.attrs["source.inference_version"] == "1.0"


@pytest.mark.parametrize(
    ["init_times", "batch_time", "expected"],
    [
        pytest.param(
            np.array([cftime.DatetimeJulian(2020, 1, 1, 0, 0, 0) for _ in range(3)]),
            np.array(
                [
                    xr.date_range(
                        cftime.DatetimeJulian(2020, 1, 1, 0, 0, 0),
                        freq="6h",
                        periods=3,
                        use_cftime=True,
                    ).values
                    for _ in range(3)
                ]
            ),
            np.array(
                [
                    MICROSECONDS_PER_SECOND * SECONDS_PER_HOUR * i
                    for i in range(0, 18, 6)
                ]
            ),
            id="init_same",
        ),
        pytest.param(
            np.array(
                [cftime.DatetimeJulian(2020, 1, 1, 6 * i, 0, 0) for i in range(3)]
            ),
            np.array(
                [
                    xr.date_range(
                        cftime.DatetimeJulian(2020, 1, 1, 6 * i, 0, 0),
                        freq="6h",
                        periods=3,
                        use_cftime=True,
                    )
                    for i in range(3)
                ]
            ),
            np.array(
                [
                    MICROSECONDS_PER_SECOND * SECONDS_PER_HOUR * i
                    for i in range(0, 18, 6)
                ]
            ),
            id="init_different",
        ),
        pytest.param(
            np.array(
                [cftime.DatetimeJulian(2020, 1, 1, 6 * i, 0, 0) for i in range(3)]
            ),
            np.array(
                [
                    xr.date_range(
                        cftime.DatetimeJulian(2020, 1, 2, 6 * i, 0, 0),
                        freq="6h",
                        periods=3,
                        use_cftime=True,
                    )
                    for i in range(3)
                ]
            ),
            np.array(
                [
                    MICROSECONDS_PER_SECOND * SECONDS_PER_HOUR * i
                    for i in range(24, 42, 6)
                ]
            ),
            id="not_initial_time_window",
        ),
    ],
)
def test_get_batch_lead_times_microseconds(init_times, batch_time, expected):
    lead_time_seconds = get_batch_lead_time_microseconds(init_times, batch_time)
    assert lead_time_seconds.shape == expected.shape
    np.testing.assert_equal(lead_time_seconds, expected)


def test_get_batch_lead_time_microseconds_length_mismatch():
    init_times = np.array(
        [cftime.DatetimeJulian(2020, 1, 1, 6 * i, 0, 0) for i in range(3)]
    )
    batch_time = np.array(
        [
            xr.date_range(
                cftime.DatetimeJulian(2020, 1, 2, 6 * i, 0, 0),
                freq="6h",
                periods=3,
                use_cftime=True,
            ).values
            for i in range(2)
        ],
    )
    with pytest.raises(ValueError):
        get_batch_lead_time_microseconds(init_times, batch_time)


def test_get_batch_lead_time_microseconds_inconsistent_samples():
    init_times = np.array(
        [cftime.DatetimeJulian(2020, 1, 1, 6, 0, 0) for _ in range(2)]
    )
    batch_time = np.array(
        [
            xr.date_range(
                cftime.DatetimeJulian(2020, 1, 1, 6, 0, 0),
                freq="6h",
                periods=3,
                use_cftime=True,
            ),
            xr.date_range(
                cftime.DatetimeJulian(2020, 1, 1, 12, 0, 0),
                freq="6h",
                periods=3,
                use_cftime=True,
            ),
        ]
    )
    with pytest.raises(ValueError):
        get_batch_lead_time_microseconds(init_times, batch_time)


@pytest.mark.parametrize(
    ["years_ahead", "overflow"],
    [
        pytest.param(1e2, False, id="100_years_OK"),
        pytest.param(1e4, False, id="10_000_years_OK"),
        pytest.param(1e6, True, id="1_000_000_years_fails"),
    ],
)
def test_get_batch_lead_time_microseconds_overflow(years_ahead, overflow):
    init_times = np.array([cftime.DatetimeNoLeap(2020, 1, 1)])
    batch_time = np.array([cftime.DatetimeNoLeap(2020 + years_ahead, 1, 1)])[:, None]
    days_per_year_noleap = 365
    seconds_per_day = 86400
    expected_lead_time_microseconds = (
        MICROSECONDS_PER_SECOND * seconds_per_day * days_per_year_noleap * years_ahead
    )
    if not overflow:
        lead_time = get_batch_lead_time_microseconds(init_times, batch_time)
        assert lead_time.item() == expected_lead_time_microseconds
    else:
        with pytest.raises(OverflowError):
            get_batch_lead_time_microseconds(init_times, batch_time)



import pytest
import torch
from torch_harmonics import InverseRealSHT

from fme.ace.registry.stochastic_sfno import isotropic_noise
from fme.core.device import get_device


@pytest.mark.parametrize("nlat, nlon", [(8, 16), (64, 128)])
def test_isotropic_noise(nlat: int, nlon: int):
    torch.manual_seed(0)
    n_batch = 1000
    embed_dim = 4
    leading_shape = (n_batch, embed_dim)
    isht = InverseRealSHT(nlat, nlon, grid="legendre-gauss")
    lmax = isht.lmax
    mmax = isht.mmax
    noise = isotropic_noise(leading_shape, lmax, mmax, isht, device=get_device())
    assert noise.shape == (n_batch, embed_dim, nlat, nlon)
    assert noise.dtype == torch.float32
    torch.testing.assert_close(
        noise.mean(), torch.tensor(0.0, device=noise.device), atol=2e-3, rtol=0.0
    )
    torch.testing.assert_close(
        noise.std(), torch.tensor(1.0, device=noise.device), atol=5e-3, rtol=0.0
    )



from .multi_call import MultiCallStep, MultiCallStepConfig
from .radiation import SeparateRadiationStep, SeparateRadiationStepConfig
from .single_module import SingleModuleStep, SingleModuleStepConfig
from .step import StepABC, StepConfigABC, StepSelector



import torch
import torch.jit

from fme.core.typing_ import TensorDict


class DataShapesNotUniform(ValueError):
    """Indicates that a set of tensors do not all have the same shape."""

    pass


class Packer:
    """
    Responsible for packing tensors into a single tensor.
    """

    def __init__(self, names: list[str]):
        self.names = names

    def pack(self, tensors: TensorDict, axis=0) -> torch.Tensor:
        """
        Packs tensors into a single tensor, concatenated along a new axis.

        Args:
            tensors: Dict from names to tensors.
            axis: index for new concatenation axis.

        Raises:
            DataShapesNotUniform: when packed tensors do not all have the same shape.
        """
        shape = next(iter(tensors.values())).shape
        for name in tensors:
            if tensors[name].shape != shape:
                raise DataShapesNotUniform(
                    f"Cannot pack tensors of different shapes. "
                    f'Expected "{shape}" got "{tensors[name].shape}"'
                )
        return _pack(tensors, self.names, axis=axis)

    def unpack(self, tensor: torch.Tensor, axis=0) -> TensorDict:
        return _unpack(tensor, self.names, axis=axis)


@torch.jit.script
def _pack(tensors: TensorDict, names: list[str], axis: int = 0) -> torch.Tensor:
    return torch.cat([tensors[n].unsqueeze(axis) for n in names], dim=axis)


@torch.jit.script
def _unpack(tensor: torch.Tensor, names: list[str], axis: int = 0) -> TensorDict:
    return {n: tensor.select(axis, index=i) for i, n in enumerate(names)}



import datetime

import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.aggregator.inference import InferenceAggregator
from fme.ace.data_loading.batch_data import PairedData
from fme.core.device import get_device

from .test_evaluator import get_ds_info

TIMESTEP = datetime.timedelta(hours=6)


def get_zero_time(shape, dims):
    return xr.DataArray(np.zeros(shape, dtype="datetime64[ns]"), dims=dims)


def test_logs_labels_exist():
    n_sample = 10
    n_time = 22
    nx = 2
    ny = 2
    ds_info = get_ds_info(nx, ny)
    agg = InferenceAggregator(ds_info, n_time, save_diagnostics=False)
    target_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    logs = agg.record_batch(
        PairedData(reference=target_data, prediction=gen_data, time=time),
    )
    assert len(logs) == n_time
    expected_step_keys = [
        "mean/forecast_step",
        "mean/weighted_mean_gen/a",
        "mean/weighted_std_gen/a",
    ]
    for log in logs:
        for key in expected_step_keys:
            assert key in log, key
        assert len(log) == len(expected_step_keys), set(log).difference(
            expected_step_keys
        )
    summary_logs = agg.get_summary_logs()
    expected_summary_keys = ["time_mean/gen_map/a", "annual/a"]
    for key in expected_summary_keys:
        assert key in summary_logs, key
    assert len(summary_logs) == len(expected_summary_keys), set(
        summary_logs
    ).difference(expected_summary_keys)


def test_logs_labels_exist_with_reference_time_means():
    n_sample = 10
    n_time = 22
    nx = 2
    ny = 2
    ds_info = get_ds_info(nx, ny)
    reference_time_means = xr.Dataset(
        {
            "a": xr.DataArray(
                np.random.randn(ny, nx).astype(np.float32),
                dims=["grid_yt", "grid_xt"],
            )
        }
    )
    agg = InferenceAggregator(
        ds_info,
        n_time,
        time_mean_reference_data=reference_time_means,
        save_diagnostics=False,
    )
    target_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    logs = agg.record_batch(
        PairedData(
            reference=target_data,
            prediction=gen_data,
            time=time,
        ),
    )
    assert len(logs) == n_time
    expected_step_keys = [
        "mean/forecast_step",
        "mean/weighted_mean_gen/a",
        "mean/weighted_std_gen/a",
    ]
    for log in logs:
        for key in expected_step_keys:
            assert key in log, key
        assert len(log) == len(expected_step_keys), set(log).difference(
            expected_step_keys
        )
    summary_logs = agg.get_summary_logs()
    expected_summary_keys = [
        "time_mean/gen_map/a",
        "time_mean/ref_bias_map/a",
        "time_mean/ref_bias/a",
        "time_mean/ref_rmse/a",
        "annual/a",
    ]
    for key in expected_summary_keys:
        assert key in summary_logs, key
    assert len(summary_logs) == len(expected_summary_keys), set(
        summary_logs
    ).difference(expected_summary_keys)


def test_flush_diagnostics(tmpdir):
    nx, ny, n_sample, n_time = 2, 2, 10, 21
    ds_info = get_ds_info(nx, ny)
    agg = InferenceAggregator(ds_info, n_time, output_dir=tmpdir)
    target_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    agg.record_batch(
        data=PairedData(
            prediction=gen_data,
            reference=target_data,
            time=time,
        ),
    )
    agg.flush_diagnostics()
    expected_files = [  # note: time-dependent aggregators not tested here
        "mean",
        "time_mean",
    ]
    for file in expected_files:
        assert (tmpdir / f"{file}_diagnostics.nc").exists()


def test_agg_raises_without_output_dir():
    ds_info = get_ds_info(nx=2, ny=2)
    with pytest.raises(
        ValueError, match="Output directory must be set to save diagnostics"
    ):
        InferenceAggregator(
            ds_info,
            n_timesteps=1,
            save_diagnostics=True,
            output_dir=None,
        )



import datetime
import logging
from collections.abc import Callable, Iterable, Iterator, Mapping, Sized
from typing import Any, Generic, Literal, TypeVar

import numpy as np
import torch
import xarray as xr

from fme.ace.data_loading.augmentation import BatchModifierABC, NullModifier
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.data_loading.dataloader import SlidingWindowDataLoader
from fme.ace.requirements import PrognosticStateDataRequirements
from fme.core.coordinates import HorizontalCoordinates, VerticalCoordinate
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset.properties import DatasetProperties
from fme.core.dataset_info import DatasetInfo
from fme.core.generics.data import DataLoader, GriddedDataABC, InferenceDataABC

T = TypeVar("T", covariant=True)


U = TypeVar("U")


class SizedMap(Generic[T, U], Sized, Iterable[U]):
    def __init__(self, func: Callable[[T], U], iterable: DataLoader[T]):
        self._func = func
        self._iterable = iterable

    def __len__(self) -> int:
        return len(self._iterable)

    def __iter__(self) -> Iterator[U]:
        return map(self._func, self._iterable)


class GriddedData(GriddedDataABC[BatchData]):
    """
    Data as required for pytorch training.

    The data is assumed to be gridded, and attributes are included for
    performing operations on gridded data.

    All data exposed from this class is on the current device.
    """

    def __init__(
        self,
        loader: DataLoader[BatchData] | SlidingWindowDataLoader,
        properties: DatasetProperties,
        modifier: BatchModifierABC = NullModifier(),
        sampler: torch.utils.data.Sampler | None = None,
    ):
        """
        Args:
            loader: torch DataLoader, which returns batches of type BatchData.
                Data can be on any device (but will typically be on CPU).
            properties: Batch-constant properties for the dataset, such as variable
                metadata and coordinate information. Data can be on any device.
            modifier: Modifier for the data loader.
            sampler: Optional sampler for the data loader. Provided to allow support for
                distributed training.

        Note:
            While input data can be on any device, all data exposed from this class
            will be on the current device.
        """
        self._loader = loader
        self._properties = properties.to_device()
        self._timestep = self._properties.timestep
        self._vertical_coordinate = self._properties.vertical_coordinate
        self._mask_provider = self._properties.mask_provider
        self._sampler = sampler
        self._modifier = modifier
        self._batch_size: int | None = None

    @property
    def loader(self) -> DataLoader[BatchData]:
        def modify_and_on_device(batch: BatchData) -> BatchData:
            return self._modifier(batch).to_device()

        return SizedMap(modify_and_on_device, self._loader)

    @property
    def variable_metadata(self) -> dict[str, VariableMetadata]:
        return self._properties.variable_metadata

    @property
    def dataset_info(self) -> DatasetInfo:
        return DatasetInfo(
            horizontal_coordinates=self.horizontal_coordinates,
            vertical_coordinate=self._vertical_coordinate,
            mask_provider=self._mask_provider,
            timestep=self._timestep,
            variable_metadata=self._properties.variable_metadata,
        )

    @property
    def horizontal_coordinates(self) -> HorizontalCoordinates:
        return self._properties.horizontal_coordinates

    @property
    def coords(self) -> Mapping[str, np.ndarray]:
        return {
            **self.horizontal_coordinates.coords,
            **self._vertical_coordinate.coords,
        }

    @property
    def grid(self) -> Literal["equiangular", "legendre-gauss", "healpix"]:
        return self.horizontal_coordinates.grid

    @property
    def n_samples(self) -> int:
        return self.n_batches * self.batch_size

    @property
    def n_batches(self) -> int:
        return len(self._loader)

    @property
    def _first_time(self) -> Any:
        return self._loader.dataset[0][1].values[0]  # type: ignore

    @property
    def _last_time(self) -> Any:
        return self._loader.dataset[-1][1].values[0]  # type: ignore

    @property
    def batch_size(self) -> int:
        if self._batch_size is None:
            example_data = next(iter(self.loader)).data
            example_tensor = next(iter(example_data.values()))
            self._batch_size = example_tensor.shape[0]
        return self._batch_size

    def log_info(self, name: str):
        if isinstance(self._loader, SlidingWindowDataLoader):
            self._loader.log_info(name)
        else:
            logging.info(
                f"{name} data: {self.n_samples} samples, {self.n_batches} batches"
            )
        logging.info(f"{name} data: first sample's initial time: {self._first_time}")
        logging.info(f"{name} data: last sample's initial time: {self._last_time}")

    def set_epoch(self, epoch: int):
        """
        Set the epoch for the data loader sampler, if it is a distributed sampler.
        """
        if self._sampler is not None and isinstance(
            self._sampler, torch.utils.data.DistributedSampler
        ):
            self._sampler.set_epoch(epoch)


def get_initial_condition(
    loader: DataLoader[BatchData],
    requirements: PrognosticStateDataRequirements,
) -> PrognosticState:
    for batch in loader:
        return batch.to_device().get_start(
            prognostic_names=requirements.names,
            n_ic_timesteps=requirements.n_timesteps,
        )
    raise ValueError("No initial condition found in loader")


class InferenceGriddedData(InferenceDataABC[PrognosticState, BatchData]):
    """
    Data as required for inference.

    All data exposed from this class is on the current device.
    """

    def __init__(
        self,
        loader: DataLoader[BatchData],
        initial_condition: PrognosticState | PrognosticStateDataRequirements,
        properties: DatasetProperties,
    ):
        """
        Args:
            loader: torch DataLoader, which returns batches of type BatchData.
                Data can be on any device (but will typically be on CPU).
            initial_condition: Initial condition for the inference, or a requirements
                object specifying how to extract the initial condition from the first
                batch of data. Data can be on any device.
            properties: Batch-constant properties for the dataset, such as variable
                metadata and coordinate information. Data can be on any device.

        Note:
            While input data can be on any device, all data exposed from this class
            will be on the current device.
        """
        self._loader = loader
        self._properties = properties.to_device()
        self._n_initial_conditions: int | None = None
        if isinstance(initial_condition, PrognosticStateDataRequirements):
            self._initial_condition: PrognosticState = get_initial_condition(
                loader, initial_condition
            )
        else:
            self._initial_condition = initial_condition.to_device()
        self._initial_time: xr.DataArray | None = None

    @property
    def loader(self) -> DataLoader[BatchData]:
        def on_device(batch: BatchData) -> BatchData:
            return batch.to_device()

        return SizedMap(on_device, self._loader)

    @property
    def variable_metadata(self) -> dict[str, VariableMetadata]:
        return self._properties.variable_metadata

    @property
    def dataset_info(self) -> DatasetInfo:
        return DatasetInfo(
            horizontal_coordinates=self.horizontal_coordinates,
            vertical_coordinate=self._vertical_coordinate,
            mask_provider=self._properties.mask_provider,
            timestep=self.timestep,
        )

    @property
    def _vertical_coordinate(self) -> VerticalCoordinate:
        return self._properties.vertical_coordinate

    @property
    def horizontal_coordinates(self) -> HorizontalCoordinates:
        return self._properties.horizontal_coordinates

    @property
    def timestep(self) -> datetime.timedelta:
        return self._properties.timestep

    @property
    def coords(self) -> Mapping[str, np.ndarray]:
        return {
            **self.horizontal_coordinates.coords,
            **self._vertical_coordinate.coords,
        }

    @property
    def n_initial_conditions(self) -> int:
        if self._n_initial_conditions is None:
            example_data = next(iter(self.loader)).data
            example_tensor = next(iter(example_data.values()))
            self._n_initial_conditions = example_tensor.shape[0]
        return self._n_initial_conditions

    @property
    def initial_condition(self) -> PrognosticState:
        return self._initial_condition

    @property
    def initial_time(self) -> xr.DataArray:
        if self._initial_time is None:
            for batch in self.loader:
                self._initial_time = batch.time.isel(time=0)
                break
            else:
                raise ValueError("No data found in loader")
        return self._initial_time


class PSType:
    pass


class FDType:
    pass


class ErrorInferenceData(InferenceDataABC[PSType, FDType]):
    """
    A inference data class that raises an error when accessed.

    Necessary because in some contexts inference is not run,
    and no data is configured (but also we don't need data).
    """

    def __init__(self):
        pass

    @property
    def initial_condition(self) -> PSType:
        raise ValueError("No inference data available")

    @property
    def loader(self) -> DataLoader[FDType]:
        raise ValueError("No inference data available")

    @property
    def initial_inference_times(self) -> xr.DataArray:
        raise ValueError("No inference data available")



# This script is used to compute a training dataset from the "raw"
# FV3GFS data stored in zarr form on GCS.

# The dependencies of this script are installed in the "fv3net" conda environment
# which can be installed using fv3net's Makefile. See
# https://github.com/ai2cm/fv3net/blob/8ed295cf0b8ca49e24ae5d6dd00f57e8b30169ac/Makefile#L310

# The resulting dataset is about 194GB (the input is about 2.5TB). Running this script
# on my 8-CPU VM takes about 2.5 hours. See "compute_dataset_fv3gfs_argo_workflow.yaml"
# for a workflow which parallelizes this script across the 11-member ensemble and runs
# it on our GKE cluster.

import os
import pdb
import sys
from typing import Tuple

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import multiprocessing as mp
from functools import partial

import click
import earth2grid
import numpy as np
import torch
import xarray as xr
import xpartition  # noqa
from compute_dataset import (
    LATENT_HEAT_OF_VAPORIZATION,
    DatasetComputationConfig,
    DatasetConfig,
    DLWPChunkingConfig,
    DLWPNameMapping,
    assert_column_integral_of_moisture_is_conserved,
    assert_global_dry_air_mass_conservation,
    assert_global_moisture_conservation,
    get_dataset_urls,
    open_datasets,
)


def regrid_tensor(x, regrid_func, shape):
    data = regrid_func(torch.tensor(x, dtype=torch.double))
    return data.numpy().reshape(shape)


def _pool_func(ds, store, n_partitions, partition_dims, i):
    ds.partition.write(store, n_partitions, partition_dims, i)
    print(f"Finished writing partition {i}")


def hpx_regrid(
    ds: xr.Dataset,
    dlwp_names: DLWPNameMapping,
    level: int = 6,  # regrid resolution
    n_side: int = 64,
) -> Tuple[xr.Dataset, xr.Dataset]:
    lat_long_names = dlwp_names.lat_lon_dims
    longitude = lat_long_names[1]
    latitude = lat_long_names[0]
    lons = ds[longitude]
    lats = ds[latitude]

    hpx = earth2grid.healpix.Grid(
        level=level, pixel_order=earth2grid.healpix.HEALPIX_PAD_XY
    )
    src = earth2grid.latlon.LatLonGrid(lat=list(lats), lon=list(lons))
    # Regridder
    regrid = earth2grid.get_regridder(src, hpx)

    ds_regridded = xr.apply_ufunc(
        regrid_tensor,
        ds,
        input_core_dims=[[latitude, longitude]],
        output_core_dims=[["face", "height", "width"]],
        output_sizes={"face": 12, "height": n_side, "width": n_side},
        output_dtypes=[float],
        dask="parallelized",
        vectorize=True,
        on_missing_core_dim="copy",
        kwargs={"regrid_func": regrid, "shape": (12, n_side, n_side)},
        dask_gufunc_kwargs={"allow_rechunk": True},
    )
    # Assign coordinates to the regridded dataset
    time_coords = ds.coords["time"]
    nside_coords = np.arange(n_side)
    grid_coords = np.arange(12)
    ds_regridded = ds_regridded.assign_coords(
        time=time_coords,
        face=grid_coords,
        height=nside_coords,
        width=nside_coords,
    )

    return ds_regridded


def construct_hpx_dataset(
    config: DatasetComputationConfig,
    run_directory: str,
    output_directory: str,
    toy_dataset: bool = False,
) -> xr.Dataset:
    dlwp_names = config.standard_names
    if not isinstance(dlwp_names, DLWPNameMapping):
        raise TypeError("Expected to be passed type of DLWPNameMapping.")
    dlwp_chunking = config.chunking
    if not isinstance(dlwp_chunking, DLWPChunkingConfig):
        raise TypeError("Expected to be passed type of DLWPChunkingConfig.")

    urls = get_dataset_urls(config, run_directory)
    print(urls)
    ds = open_datasets(config, urls)
    for var in ds:
        del ds[var].encoding["chunks"]
        del ds[var].encoding["preferred_chunks"]
    print(f"Input dataset size is {ds.nbytes / 1e9} GB")
    if toy_dataset:
        ds = ds.isel(time=slice(0, 200))
    # We would like to:

    # 1. map to healpix mesh
    ds = hpx_regrid(
        ds=ds,
        dlwp_names=dlwp_names,
        n_side=64,
    )
    print(f"After regrid: {ds}")

    # 2. chunk and save
    chunks = config.chunking.get_chunks(dlwp_names)
    ds = ds.chunk(chunks)
    ds.attrs["history"] = (
        "Dataset computed by full-model/scripts/data_process"
        "/compute_hpx_dataset.py"
        f" script, using following input zarrs: {urls.values()}."
    )
    ds.attrs["vertical_coordinate"] = (
        "The pressure at level interfaces can by computed as "
        "p_i = ak_i + bk_i * PRESsfc, where PRESsfc is the surface pressure and the "
        "p_i pressure corresponds to the interface at the top of the i'th finite "
        "volume layer, counting down from the top of atmosphere."
    )
    ds = ds.rename(config.renaming)
    return ds


@click.command()
@click.option("--config", help="Path to dataset configuration YAML file.")
@click.option("--run-directory", help="Path to reference run directory.")
@click.option("--output-store", help="Path to output zarr store.")
@click.option("--debug", is_flag=True, help="Print metadata instead of writing output.")
@click.option("--subsample", is_flag=True, help="Subsample the data before writing.")
@click.option("--check-conservation", is_flag=True, help="Check conservation.")
@click.option("--num-processes", default=16, help="Number of processes to spin up.")
def main(
    config,
    run_directory,
    output_store,
    debug,
    subsample,
    check_conservation,
    num_processes,
):
    config = DatasetConfig.from_file(config).dataset_computation
    dlwp_names = config.standard_names
    print(f"--run-directory is {run_directory}")
    print(f"--output-store is {output_store}")
    ds = construct_hpx_dataset(
        config=config,
        run_directory=run_directory,
        output_directory=output_store,
        toy_dataset=False,
    )
    if subsample:
        ds = ds.isel(time=slice(10, 13))
    if check_conservation:
        assert_column_integral_of_moisture_is_conserved(
            ds,
            precipitable_water_path_name=dlwp_names.precipitable_water_path,
            total_water_path_name=dlwp_names.total_water_path,
        )
        assert_global_dry_air_mass_conservation(
            ds,
            dims=dlwp_names.horizontal_dims,
            surface_pressure_name=dlwp_names.surface_pressure,
            total_water_path_name=dlwp_names.total_water_path,
            latitude_dim=dlwp_names.latitude_dim,
            time_dim=dlwp_names.time_dim,
        )
        assert_global_moisture_conservation(
            ds,
            dims=dlwp_names.horizontal_dims,
            latitude_dim=dlwp_names.latitude_dim,
            total_water_path_name=dlwp_names.total_water_path,
            latent_heat_flux_name=dlwp_names.latent_heat_flux,
            latent_heat_of_vaporization=LATENT_HEAT_OF_VAPORIZATION,
            precip_rate_name=dlwp_names.precip_rate,
            time_dim=dlwp_names.time_dim,
        )
    drop_vars = [var for var in dlwp_names.dropped_variables if var in ds]
    ds = ds.drop(drop_vars)
    print(f"Output dataset size is {ds.nbytes / 1e9} GB")

    if debug:
        with xr.set_options(display_max_rows=500):
            print(ds)
    else:
        n_partitions = config.n_split
        partition_dims = [dlwp_names.time_dim]
        store = f"{output_store}.zarr"
        ds.partition.initialize_store(store)

        with mp.get_context("forkserver").Pool(num_processes) as pool:
            pool.map(
                partial(_pool_func, ds, store, n_partitions, partition_dims),
                range(n_partitions),
            )


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)
        pdb.post_mortem()  # Start the debugger
        raise  # Re-raise the exception to preserve the traceback



import dataclasses
import datetime

import torch

from fme.core.typing_ import TensorDict, TensorMapping

from .atmosphere_data import AtmosphereData
from .constants import DENSITY_OF_WATER, SPECIFIC_HEAT_OF_WATER
from .prescriber import Prescriber


@dataclasses.dataclass
class SlabOceanConfig:
    """
    Configuration for a slab ocean model.

    Parameters:
        mixed_layer_depth_name: Name of the mixed layer depth field.
        q_flux_name: Name of the heat flux field.
    """

    mixed_layer_depth_name: str
    q_flux_name: str

    @property
    def names(self) -> list[str]:
        return [self.mixed_layer_depth_name, self.q_flux_name]


@dataclasses.dataclass
class OceanConfig:
    """
    Configuration for determining sea surface temperature from an ocean model.

    Parameters:
        surface_temperature_name: Name of the sea surface temperature field.
        ocean_fraction_name: Name of the ocean fraction field.
        interpolate: If True, interpolate between ML-predicted surface temperature and
            ocean-predicted surface temperature according to ocean_fraction. If False,
            only use ocean-predicted surface temperature where ocean_fraction>=0.5.
        slab: If provided, use a slab ocean model to predict surface temperature.
    """

    surface_temperature_name: str
    ocean_fraction_name: str
    interpolate: bool = False
    slab: SlabOceanConfig | None = None

    def build(
        self,
        in_names: list[str],
        out_names: list[str],
        timestep: datetime.timedelta,
    ) -> "Ocean":
        if not (
            self.surface_temperature_name in in_names
            and self.surface_temperature_name in out_names
        ):
            raise ValueError(
                "To use a surface ocean model, the surface temperature must be present"
                f" in_names and out_names, but {self.surface_temperature_name} is not."
            )
        return Ocean(config=self, timestep=timestep)

    @property
    def forcing_names(self) -> list[str]:
        names = [self.ocean_fraction_name]
        if self.slab is None:
            names.append(self.surface_temperature_name)
        else:
            names.extend(self.slab.names)
        return list(set(names))


class Ocean:
    """Overwrite sea surface temperature with that predicted from some ocean model."""

    def __init__(self, config: OceanConfig, timestep: datetime.timedelta):
        """
        Args:
            config: Configuration for the surface ocean model.
            timestep: Timestep of the model.
        """
        self.surface_temperature_name = config.surface_temperature_name
        self.ocean_fraction_name = config.ocean_fraction_name
        self.prescriber = Prescriber(
            prescribed_name=config.surface_temperature_name,
            mask_name=config.ocean_fraction_name,
            mask_value=1,
            interpolate=config.interpolate,
        )
        self._forcing_names = config.forcing_names
        if config.slab is None:
            self.type = "prescribed"
        else:
            self.type = "slab"
            self.mixed_layer_depth_name = config.slab.mixed_layer_depth_name
            self.q_flux_name = config.slab.q_flux_name
        self.timestep = timestep

    def __call__(
        self,
        input_data: TensorMapping,
        gen_data: TensorMapping,
        target_data: TensorMapping,
    ) -> TensorDict:
        """
        Args:
            input_data: Denormalized input data for current step.
            gen_data: Denormalized output data for current step.
            target_data: Denormalized data that includes mask and forcing data. Assumed
                to correspond to the same time step as gen_data.

        Returns:
            gen_data with sea surface temperature overwritten by ocean model.
        """
        if self.type == "prescribed":
            next_step_temperature = target_data[self.surface_temperature_name]
        elif self.type == "slab":
            temperature_tendency = mixed_layer_temperature_tendency(
                AtmosphereData(gen_data).net_surface_energy_flux_without_frozen_precip,
                target_data[self.q_flux_name],
                target_data[self.mixed_layer_depth_name],
            )
            next_step_temperature = (
                input_data[self.surface_temperature_name]
                + temperature_tendency * self.timestep.total_seconds()
            )
        else:
            raise NotImplementedError(f"Ocean type={self.type} is not implemented")

        return self.prescriber(
            target_data,
            gen_data,
            {self.surface_temperature_name: next_step_temperature},
        )

    @property
    def forcing_names(self) -> list[str]:
        """These are the variables required from the forcing data."""
        return self._forcing_names


def mixed_layer_temperature_tendency(
    f_net: torch.Tensor,
    q_flux: torch.Tensor,
    depth: torch.Tensor,
    density=DENSITY_OF_WATER,
    specific_heat=SPECIFIC_HEAT_OF_WATER,
) -> torch.Tensor:
    """
    Args:
        f_net: Net surface energy flux in W/m^2.
        q_flux: Convergence of ocean heat transport in W/m^2.
        depth: Mixed layer depth in m.
        density (optional): Density of water in kg/m^3.
        specific_heat (optional): Specific heat of water in J/kg/K.

    Returns:
        Temperature tendency of mixed layer in K/s.
    """
    return (f_net + q_flux) / (density * depth * specific_heat)



import dataclasses
from collections.abc import Mapping
from typing import Any

from fme.ace.models.ocean.m2lines.samudra import Samudra
from fme.ace.registry.registry import ModuleConfig, ModuleSelector


@ModuleSelector.register("Samudra")
@dataclasses.dataclass
class SamudraBuilder(ModuleConfig):
    """
    Configuration for the M2Lines Samudra architecture.
    """

    ch_width: list[int] = dataclasses.field(
        default_factory=lambda: [200, 250, 300, 400]
    )
    n_layers: list[int] = dataclasses.field(default_factory=lambda: [1, 1, 1, 1])
    dilation: list[int] = dataclasses.field(default_factory=lambda: [1, 2, 4, 8])
    pad: str = "circular"
    norm: str = "instance"
    norm_kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        if "num_features" in self.norm_kwargs:
            raise ValueError("norm_kwargs should not have num_features")
        if "normalized_shape" in self.norm_kwargs:
            raise ValueError("norm_kwargs should not have normalized_shape")

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ):
        return Samudra(
            input_channels=n_in_channels,
            output_channels=n_out_channels,
            ch_width=self.ch_width,
            dilation=self.dilation,
            n_layers=self.n_layers,
            pad=self.pad,
            norm=self.norm,
            norm_kwargs=self.norm_kwargs,
        )



import dataclasses


@dataclasses.dataclass
class PrognosticStateDataRequirements:
    """
    The requirements for the model's prognostic state.

    Parameters:
        names: Names of prognostic variables.
        n_timesteps: Number of consecutive timesteps that must be stored.
    """

    names: list[str]
    n_timesteps: int


@dataclasses.dataclass
class DataRequirements:
    """
    ACE's requirements for batches (time windows) of loaded data.

    Parameters:
        names: Names of the variables to load.
        n_timesteps: Number of timesteps to load in each batch window.
    """

    names: list[str]
    n_timesteps: int


NullDataRequirements = DataRequirements(
    names=[],
    n_timesteps=0,
)
